/*
 * xsk.c --- sk.c + skcache.c
 */

/*
 * sk.c --- skip list procedures
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 */

/*
 * skcache.c --- object cache procedures using skip list and its bucket
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 *
 * $Id$
 */

/*
 * See also:
 *   William Pugh, "skiplist cookbook",
 *   UMIACS-TR-89-72.1, July 1989
 */

/*
 * Original algorithm was described by Pascal like langauage.
 * Subscription were 1..maxlevel or 1..level .
 * Since this is C program,  This file describes subscription as
 * 0..maxlevel-1 and 0..level-1,
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <dirent.h>

#include <malloc.h>
#include <assert.h>
#include <errno.h>

#include "range.h"
#include "msg.h"

#include "xsk.h"
#include "sim_stat.h"

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif
#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

/*
 * arround 256, follwing numbers are known as prime numbers.
 *  241 251 257 263 269 271
 *
 */

extern int radix_h1;
extern int radix_h2;
extern int radix_h3;
extern int radix_h4;

extern char* cache_dir_prefix;

#ifndef HAVE_DIRENT_T
typedef struct dirent dirent_t;
#endif

int
rm_fr(char *base)
{
    char tmp[L_PATH];
    dirent_t    *dp;
    DIR         *dirp;
#ifdef PC_NAME_MAX
    dirent_t ent[PC_NAME_MAX+1];
#else
#ifdef _PC_NAME_MAX
    dirent_t ent[_PC_NAME_MAX+1];
#else
    dirent_t ent[128];
#endif
#endif
    struct stat sbuf;
    int chk;

    if(base==NULL)
        return -1;

#if 0
Trace("rm_fr: called '%s'\n", base);
#endif

    chk = rmdir(base);
    if(chk) {
        if(errno==ENOTEMPTY) {

            dirp = opendir(base);
            if(dirp) {
                while(1) {
                    chk = readdir_r(dirp, ent, &dp);
                    if(dp==NULL) {
                        break;
                    }

                    if(dp->d_name[0]=='.' && dp->d_name[1]=='\0') {
                        continue;
                    }
                    if(dp->d_name[0]=='.'
                        && dp->d_name[1]=='.' && dp->d_name[2]=='\0') {
                        continue;
                    }
                    strcpy(tmp, base);
                    strcat(tmp, "/");
                    strcat(tmp, dp->d_name);


                    chk = stat(tmp, &sbuf);
                    if(chk) {
                        Error("rm_fr: fail stat() '%s' (%d)\n",
                            tmp, errno);
                        continue;
                    }

                    if(S_ISREG(sbuf.st_mode)) {
                        unlink(tmp);
                    }
                    if(S_ISDIR(sbuf.st_mode)) {
                        rm_fr(tmp);
                    }
                }
                closedir(dirp);
            }

            chk = rmdir(base);

            if(chk) {
                Error("rm_fr: fail rmdir() '%s' after sweep (%d)\n",
                    base, errno);
            }

        }
    }

	return 0;
}

void
Banker(void* dummy)
{
    int d1, d2, d3;
    char tmp[L_PATH];
    char path1[L_PATH];
    char path2[L_PATH];
    char path3[L_PATH];
    char path4[L_PATH];
    struct stat sbuf;
    off_t sum_f_count;
    off_t sum_f_size;
    off_t tmp_f_size;
    time_t begin, end;
    int chk;

    dirent_t    *dp;
    DIR         *dirp;
#ifdef PC_NAME_MAX
    dirent_t ent[PC_NAME_MAX+1];
#else
#ifdef _PC_NAME_MAX
    dirent_t ent[_PC_NAME_MAX+1];
#else
    dirent_t ent[128];
#endif
#endif

    int level_interval = 10;
    int loop_interval = 3600;


  while(1) {

    begin = time(NULL);
    Trace("Banker wakeup at %d\n", begin);

    sum_f_size = (off_t) 0;
    sum_f_count = (off_t) 0;

    for(d1=0;d1<radix_h1;d1++) {
        Trace("banker %d\n", d1);
        sprintf(path1, "%s/%d", cache_dir_prefix, d1);

        chk = stat(path1, &sbuf);
        if(chk) {
            if(errno==ENOENT) {
                chk = mkdir(path1, 0755);
                if(chk) {
                    Error("level 1: fail mkdir() '%s' (%d)\n", path1, errno);
                    break;
                }
            }
            else {
                Error("level 1: fail stat() '%s' (%d)\n", path1, errno);
                break;
            }
        }

        for(d2=0;d2<radix_h2;d2++) {
            strcpy(path2, path1);
            sprintf(tmp, "/%d", d2);
            strcat(path2, tmp);

            chk = stat(path2, &sbuf);
            if(chk) {
                if(errno==ENOENT) {
                    chk = mkdir(path2, 0755);
                    if(chk) {
                        Error("level 2: fail mkdir() '%s' (%d)\n",
                            path2, errno);
                        break;
                    }
                }
                else {
                    Error("level 2: fail stat() '%s' (%d)\n",
                        path2, errno);
                    break;
                }
            }

            for(d3=0;d3<radix_h3;d3++) {
                strcpy(path3, path2);
                sprintf(tmp, "/%d", d3);
                strcat(path3, tmp);

                chk = stat(path3, &sbuf);
                if(chk) {
                    if(errno==ENOENT) {
                        chk = mkdir(path3, 0755);
                        if(chk) {
                            Error("level 3: fail mkdir() '%s' (%d)\n",
                                path3, errno);
                            break;
                        }
                    }
                    else {
                        Error("level 3: fail stat() '%s' (%d)\n",
                            path3, errno);
                        break;
                    }
                }
                else {

                    dirp = opendir(path3);
                    if(dirp) {
                        while(1) {
                            chk = readdir_r(dirp, ent, &dp);
                            if(dp==NULL) {
                                break;
                            }

                            if(dp->d_name[0]=='.') {
                                continue;
                            }
                            strcpy(path4, path3);
                            strcat(path4, "/");
                            strcat(path4, dp->d_name);

                            sum_f_count++;

                            chk = stat(path4, &sbuf);
                            if(chk) {
                                Error("level 4: fail stat() '%s' (%d)\n",
                                    path4, errno);
                                tmp_f_size = -1;
                            }
                            else {
                                tmp_f_size = sbuf.st_size;
                                sum_f_size += tmp_f_size;
                            }
#if 1
                            Trace("%7d: %s %9ld %12lld\n",
                                sum_f_count, path4,
                                (long)tmp_f_size, (long long)sum_f_size);
#endif
                        }
                        closedir(dirp);
                    }
                }
            }

            dirp = opendir(path2);
            if(dirp) {
                while(1) {
                    chk = readdir_r(dirp, ent, &dp);
                    if(dp==NULL) {
                        break;
                    }

                    if(dp->d_name[0]=='.') {
                        continue;
                    }
                    strcpy(path3, path2);
                    strcat(path3, "/");
                    strcat(path3, dp->d_name);

                    chk = stat(path3, &sbuf);
                    if(!chk) {
                        if(S_ISREG(sbuf.st_mode)) {
                            chk = unlink(path3);
#if 1
    Trace("2 remove r '%s' -> %d (%d)\n", path3, chk, errno);
#endif
                        }
                        else
                        if(S_ISDIR(sbuf.st_mode)) {
                            d3 = atoi(dp->d_name);
                            if(strcmp(dp->d_name, "0")!=0
                                && (d3 < 0 || d3>=radix_h3) ) {
#if 0
                            chk = rmdir(path3);
#endif
                            chk = rm_fr(path3);
#if 1
    Trace("2 remove d '%s' -> %d (%d)\n", path3, chk, errno);
#endif
                            }
                        }

                    }
                }
                closedir(dirp);
            }

            if(d2%4==0) {
                sleep(level_interval);
            }
        }

        dirp = opendir(path1);
        if(dirp) {
            while(1) {
                chk = readdir_r(dirp, ent, &dp);
                if(dp==NULL) {
                    break;
                }
                if(dp->d_name[0]=='.') {
                    continue;
                }
                strcpy(path2, path1);
                strcat(path2, "/");
                strcat(path2, dp->d_name);

                chk = stat(path2, &sbuf);
                if(!chk) {
                    if(S_ISREG(sbuf.st_mode)) {
                        chk = unlink(path2);
#if 1
    Trace("1 remove r '%s' -> %d (%d)\n", path2, chk, errno);
#endif
                    }
                    else
                    if(S_ISDIR(sbuf.st_mode)) {
                        d2 = atoi(dp->d_name);
                        if(strcmp(dp->d_name, "0")!=0
                            && (d2 < 0 || d2>=radix_h2) ) {
#if 0
                        chk = rmdir(path2);
#endif
                        chk = rm_fr(path2);

#if 1
    Trace("1 remove d '%s' -> %d (%d)\n", path2, chk, errno);
#endif
                        }
                    }


                }
            }
            closedir(dirp);
        }

        sleep(level_interval);
    }

    dirp = opendir(cache_dir_prefix);
    while(1) {
        chk = readdir_r(dirp, ent, &dp);
        if(dp==NULL) {
            break;
        }
        if(dp->d_name[0]=='.') {
            continue;
        }
        strcpy(path1, cache_dir_prefix);
        strcat(path1, "/");
        strcat(path1, dp->d_name);

        chk = stat(path1, &sbuf);
        if(!chk) {
            if(S_ISREG(sbuf.st_mode)) {
                chk = unlink(path1);
#if 1
Trace("0 remove '%s' -> %d (%d)\n", path1, chk, errno);
#endif
            }
        }
    }
    closedir(dirp);

    end = time(NULL);

    Trace("banker walking takes %d seconds\n", end-begin);
    Trace("  have %d %dx%dx%d, directories \n",
        radix_h1 * radix_h2 * radix_h3, radix_h1, radix_h2, radix_h3);
    Trace("  have %ld files (%lld bytes)\n",
        (long)sum_f_count, (long long)sum_f_size);

    sleep(loop_interval);
  }

    /*NOTREACHED*/
    exit(123);
}


