/*
 * skcache.c - object cache procedures using skip list and its bucket
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 *
 * $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <limits.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "range.h"
#include "msg.h"
#include "sk.h"

#include "skcache.h"

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif
#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif



skcache_entry_t *NIL;
int skcache_bucket_num;
bucket_t *skcache_bucket_base;
skcache_entry_t **skcache_entry_freelist_base;
skcache_frag_t   **skcache_frag_freelist_base;




/* used for index */
static int prime_box[] = {23,19,13,31,11,67,7,71,5,3,29};

int
first_hash(unsigned char *b)
{
    int l;
    register int sum, c;
    register unsigned char *p, *r, *s;

    sum = 0;

    c = 1;
    p = b;
    while(*p) {
        sum += *p - ' ' + c;
        if(*p>='0' && *p<='9') {
            /* sum += (*p - '0' + 3) * 3; */
            sum += prime_box[*p - '0'] * c;
        }
        p++;
        c++;
    }

    if(sum<0)
        sum = -sum;

/*
    return sum % nblock;
*/
    return sum % skcache_bucket_num;
}

int
second_hash(unsigned char *b)
{
    register int sum;
    register unsigned char *p;
    register int c;

    sum = 0;
    c = 1;

    p = b;
    while(*p) {
        sum = sum<<2;
        sum += prime_box[c%10]*prime_box[(c* (*p))%10];
        c++;
        p++;
    }

    if(sum<0)
        sum = -sum;

    return sum % 4099;
}

int
third_hash(unsigned char *b)
{
    register int sum;
    register unsigned char *p;
    register int c;

    sum = 0;
    c = 1;

    p = b;
    while(*p) {
        sum = sum<<2;
        sum += c * (*p) * prime_box[(c* (*p))%10];
        c++;
        p++;
    }

    if(sum<0)
        sum = -sum;

    return sum % 9997;
}








int
skcache_entry_cmp(void *px, void *py)
{
    skcache_entry_t *x = (skcache_entry_t*)px;
    skcache_entry_t *y = (skcache_entry_t*)py;

    if(x->h1==y->h1) {
        if(x->h2==y->h2) {
            return x->h3 - y->h3;
        }
        else {
            return x->h2 - y->h2;
        }
    }
    else {
        return x->h1 - y->h1;
    }
}

char*
skcache_label_hash(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_PATH];

    sprintf(tmp, "%04d:%04d:%04d", p->h1, p->h2, p->h3);

    return tmp;
}

char*
skcache_label_short(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d %s", p->h1, p->h2, p->h3, p->name);

    return tmp;
}

char*
skcache_label_ref(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d-%03d", p->h1, p->h2, p->h3, p->ref);

    return tmp;
}




int
skcache_have(int fh, int sh, int th, char *name, int flag_check_rcode, time_t ref_time)
{
    int ret;
    skcache_entry_t tmp;
    skcache_entry_t *stored_pos;
    bucket_t *bk;


    ret = 0;

    bk = &skcache_bucket_base[fh];

    tmp.h1 = fh;
    tmp.h2 = sh;
    tmp.h3 = th;
    strcpy(tmp.name, name);

    MUTEX_LOCK(&bk->mutex);

    stored_pos = (skcache_entry_t*) sk_find(&bk->cup, &tmp, skcache_entry_cmp);

    if(stored_pos) {
#if 1
        if(strcmp((const char*)stored_pos->name, (const char*)name)!=0) {
            Error("skcache_have: hash collision %04d:%04d:%04d '%s' v.s. '%s'",
                fh, sh, th, stored_pos->name, name);
            goto out;
        }
#endif

        if(flag_check_rcode) {
        }

        if(ref_time>0) {
        }

        ret = 1;
    }
    else {
        ret = 0;
    }

out:
    MUTEX_UNLOCK(&bk->mutex);

    return ret;
}


static
int
alloc_cache_entry_freelist(skcache_entry_t** head, int num)
{
    char *base;
    int esize;
    int i;
    int ret;
    skcache_entry_t *pos;
    skcache_entry_t *last;

    esize = sizeof(skcache_entry_t);
    base = (char*)malloc(esize * num);
    if(base==NULL) {
        return 0;
    }

    for(i=0; i<num; i++) {
        pos = (skcache_entry_t*)(base + esize*i);

        pos->h1 = -1;
        pos->h2 = -1;
        pos->h3 = -1;
        pos->oid = -1;
        pos->name[0] = '\0';
        pos->m_recorded = NULL;
        pos->d_recorded = -1;

        pos->next = (skcache_entry_t*)((char*)pos + esize);
    }

    *head = (skcache_entry_t*)base;
    pos->next = NULL;

#if 0

    Trace("entry// head %p, num %d\n", head, num);
    Trace("esize %d, base %p\n", esize, base);

    if(num<100) {
    i = 0;
    pos = *head;
    while(pos) {
        Trace(" %2d: pos %p, name '%s', m_recorded %p ->next %p\n",
            i, pos, pos->name, pos->m_recorded, pos->next);

        pos = pos->next;
        i++;
    }
    }

#endif

    return esize*num;
}

static
int
alloc_object_frag_freelist(skcache_frag_t** head, int num, int sz)
{
    char *base;
    int esize;
    int i;
    skcache_frag_t *pos;
    skcache_frag_t *last;

    esize = sizeof(skcache_frag_t)+sz;
    base = (char*)malloc(esize * num);
    if(base==NULL) {
        return 0;
    }

    for(i=0; i<num; i++) {
        pos = (skcache_frag_t*)(base + esize*i);
        pos->size = sz;
        pos->len = 0;
        pos->ent = (char*)((char*)pos+sizeof(skcache_frag_t));

        pos->next = (skcache_frag_t*)((char*)pos+esize);
    }
    *head = (skcache_frag_t*)base;
    pos->next = NULL;

#if 0

    Trace("frag// head %p, num %d, sz %d\n", head, num, sz);
    Trace("esize %d, base %p\n", esize, base);

    if(num<100) {
        i = 0;
        pos = *head;
        while(pos) {
            Trace(" %2d: pos %p, len %d, sz %d ->next %p\n",
                i, pos, pos->len, pos->size, pos->next);

            pos = pos->next;
            i++;
        }
    }

#endif

    return esize*num;
}



int
skcache_entry_purge(skcache_entry_t *e, int v)
{
    skcache_frag_t *cur, *pre;

#if 0
    Trace("skcache_entry_purge: p %p\n", e);
#endif

    if(e->m_recorded) {
        /* remove */
        cur = e->m_recorded;
        while(cur) {
            pre = cur->next;

            cur->next = skcache_frag_freelist_base[v];
            skcache_frag_freelist_base[v] = cur;

            cur = pre;
        }

        e->m_recorded = NULL;
    }

    /* XXX */
#if 0
    if(e->d_recorded) {
    }
#endif

    /* remove */
    e->next = skcache_entry_freelist_base[v];
    skcache_entry_freelist_base[v] = e;

    return 0;
}

char* cache_dir_prefix = "/var/tmp/cache";


int
skcache_entry_swapout(skcache_entry_t *e, int v)
{
    int sfd;
    char name[L_PATH];
    skcache_frag_t *cur, *pre;
    int nw;
    off_t nw_sum;

#if 1
    Trace("skcache_entry_swapout: p %p\n", e);
#endif
    if(e->d_recorded) {
        Error("already recorded to disk, escape");
        goto out;
    }

    sprintf(name, "%s/%d/%d/%d/%d\n",
        cache_dir_prefix, e->h1, e->h2, e->h3, e->oid);
#if 1
    Trace("  name '%s'\n", name);
#endif
    sprintf(name, "%s/%d-%d-%d-%d\n",
        cache_dir_prefix, e->h1, e->h2, e->h3, e->oid);
#if 1
    Trace("  name '%s'\n", name);
#endif
    sfd = open(name, O_TRUNC | O_CREAT, 0777);
    if(sfd<0) {
        Error("cannot open swap file");
        goto out;
    }

    nw_sum = 0L;
    if(e->m_recorded) {
        cur = e->m_recorded;
        while(cur) {
            nw = write(sfd, cur->ent, cur->len);
            if(nw>0) {
                nw_sum += nw;
            }

            cur = cur->next;
        }
        Trace("  swap out %ld bytes\n", (long)nw_sum);

        cur = e->m_recorded;
        while(cur) {
            pre = cur->next;

            cur->next = skcache_frag_freelist_base[v];
            skcache_frag_freelist_base[v] = cur;

            cur = pre;
        }

        e->m_recorded = NULL;

        e->d_recorded = e->oid;
    }
    close(sfd);

out:
    return 0;
}

int
skcache_entry_purge_check_noref(void *pp, void *vv)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    int *v = (int*)vv;
    int ret = 0;

#if 0
    Trace("  entry_check_noref: p %p ref %d in bk# %d ", p, p->ref, *v);
#endif

    if(p->ref>0) {
#if 0
        Trace("refered %d session, not purge\n", p->ref);
#endif
        ret = 0;
        goto skip;
    }
    else {
#if 0
        Trace("no refered , *purge*\n");
#endif
        skcache_entry_purge(p, *v);

        ret = 1;
    }

skip:
    return ret;

}


/*
pthread_mutex_t skcache_config_lock;
*/
extern pthread_mutex_t config_lock;

int skcache_keep_min = 30;
int skcache_keep_cur = 3600;
int skcache_keep_max = 3600;


int
skcache_entry_purge_check(void *pp, void *vv)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    int *v = (int*)vv;
    int ret = 0;
    time_t now;
    int elap, idle;
    int erase;

    erase = 0;

    now = time(NULL);
    if(p->issue_time > 0 ) {
        elap = now - p->issue_time;
    }
    else {
#if 0
        Trace("empty issue_time\n");
#endif
        elap = skcache_keep_max;
    }
    if(p->hot > 0 ) {
        idle = now - p->hot;
    }
    else {
#if 0
        Trace("empty hot\n");
#endif
        idle = skcache_keep_cur/2;
    }

    if(p->ref>0) {
        erase = 0;
    }
    else
    if(p->expire>0 && p->expire >= now) {
        erase = 1;
    }
    else
    if(idle >= skcache_keep_cur) {
        erase = 1;
    }

#if 0
    Trace("  check %p,%d ref %d, freq %d, idle %4d/%d, elap %5d: ",
        p, *v, p->ref, p->count, idle, skcache_keep_cur, elap);
#endif

    if(erase>0) {
#if 0
        Trace("*purge*\n");
#endif
        skcache_entry_purge(p, *v);
        ret = 1;
    }
    else {
#if 0
        Trace("not purge\n");
#endif
        ret = 0;
    }

skip:
    return ret;

}

int
skcache_entry_swapout_check(void *pp, void *vv)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    int *v = (int*)vv;
    int ret = 0;
    time_t now;
    int elap, idle;
    int erase;

    erase = 0;

    now = time(NULL);
    if(p->issue_time > 0 ) {
        elap = now - p->issue_time;
    }
    else {
#if 0
        Trace("empty issue_time\n");
#endif
        elap = skcache_keep_max;
    }
    if(p->hot > 0 ) {
        idle = now - p->hot;
    }
    else {
#if 0
        Trace("empty hot\n");
#endif
        idle = skcache_keep_cur/2;
    }

    if(p->ref>0) {
        erase = 0;
    }
    else
    if(p->expire>0 && p->expire >= now) {
        erase = 1;
    }
    else
    if(idle >= skcache_keep_cur) {
        erase = 1;
    }

#if 0
    Trace("  check %p,%d ref %d, freq %d, idle %4d/%d, elap %5d: ",
        p, *v, p->ref, p->count, idle, skcache_keep_cur, elap);
#endif

    if(erase>0) {
#if 0
        Trace("*swapout*\n");
#endif
        skcache_entry_swapout(p, *v);
        ret = 0;
    }
    else {
#if 0
        Trace("not purge\n");
#endif
        ret = 0;
    }

skip:
    return ret;

}



#define GC_LI   (2)

int
skcache_compaction(bucket_t *bk, int v, int force)
{
    int count;
    int deleted;
    int sum_deleted;
    skcache_entry_t *mark;
    int n_before, n_after;
    int hwater, lwater;
    double hrate, lrate;
    struct timeval t_before, t_after;
    int t_diff;

#if 1
    Trace("= = = \n");
    Trace("skcache_compaction: bk %p %s\n",
        bk, force==1? "*force*" : "normal");
#endif

    mark = skcache_entry_freelist_base[v];


#if 0
    sk_rawdump(stderr, &bk->cup, skcache_label_ref);
#endif

  hwater = 4*bk->elemnum/5;
  hrate  = 9.0/10.0;
  lwater = 3*bk->elemnum/5;
  lrate  = 10.0/9.0;

  sum_deleted = 0L;

  Log("; GC is called");

  count = 0;
  do {

    n_before = bk->cup.num;

    gettimeofday(&t_before, NULL);
    if(force==1) {
        deleted = sk_mapdelete(&bk->cup, &bk->freelist,
                skcache_entry_purge_check_noref, (void*)&v);
    }
    else {
#if 0
        deleted = sk_mapdelete(&bk->cup, &bk->freelist,
                skcache_entry_purge_check, (void*)&v);
#endif
        deleted = sk_mapdelete(&bk->cup, &bk->freelist,
                skcache_entry_swapout_check, (void*)&v);

    }
    gettimeofday(&t_after, NULL);

    n_after = bk->cup.num;

    t_diff = KOTETU_SUB_TIME(t_after, t_before);

#if 1
    Trace("  bucket# %d <%p> [%d]\n", v, bk, count);
    Trace("    elemnum %d (high water %d, low water %d)\n",
        bk->elemnum, hwater, lwater);
    Trace("    before  %d (%7.2f%%)\n",
        n_before, (double)n_before*100.0/bk->elemnum);
    Trace("    after   %d (%7.2f%%)\n",
        n_after, (double)n_after*100.0/bk->elemnum);
    Trace("    deleted %d items\n", deleted);
    Trace("    take %d msec\n", t_diff);
#endif

    if(n_after >= hwater) {
        MUTEX_LOCK(&config_lock);
        skcache_keep_cur = MAX((int)(hrate*skcache_keep_cur), skcache_keep_min);
        MUTEX_UNLOCK(&config_lock);
    }
    else
    if(n_after <= lwater) {
        MUTEX_LOCK(&config_lock);
        skcache_keep_cur = MIN((int)(lrate*skcache_keep_cur), skcache_keep_max);
        MUTEX_UNLOCK(&config_lock);
    }

#if 1
    Trace("  keeptime %d < %d < %d\n",
        skcache_keep_min, skcache_keep_cur, skcache_keep_max);
#endif

    sum_deleted += deleted;

    count++;
  } while(count<GC_LI && n_before == n_after);


#if 0
    sk_rawdump(stderr, &bk->cup, skcache_label_ref);
#endif

#if 0
    Trace("  bucket# %d\n", v);
    Trace("    bucket %p\n", &skcache_bucket_base[v]);
    Trace("    entry  %p\n", skcache_entry_freelist_base[v]);
    Trace("    frag   %p\n", skcache_frag_freelist_base[v]);
#endif

#if 0
 {
    skcache_entry_t *pos;
    int i;

    Trace("entry:\n");
    i = 0;
    pos = skcache_entry_freelist_base[v];
    while(pos) {
#if 0
        if(i<3) {
        Trace(" %2d: pos %p, name '%s', m_recorded %p ->next %p\n",
            i, pos, pos->name, pos->m_recorded, pos->next);
        }
#endif
        Trace(" %2d: pos %p, name '%s', m_recorded %p ->next %p\n",
            i, pos, pos->name, pos->m_recorded, pos->next);
        if(pos==mark) {
            Trace("     *marked*\n");
        }

        pos = pos->next;
        i++;
    }
    Trace(" ..num %d\n", i);
 }
#endif
#if 0
 {
    skcache_frag_t *pos;
    int i;

    Trace("frag:\n");
    i = 0;
    pos = skcache_frag_freelist_base[v];
    while(pos) {
        if(i<7) {
        Trace(" %2d: pos %p, len %d, sz %d ->next %p\n",
            i, pos, pos->len, pos->size, pos->next);
        }

        pos = pos->next;
        i++;
    }
    Trace(" ..num %d\n", i);
 }
#endif


#if 0
    Trace("  ret %d\n", c);
#endif

    return sum_deleted;
}



int
skcache_calc_param(long long want,
    int *xbk_num, int *xsk_level, int *xsk_len, int *xffl_num, int *xffl_size)
{
#if 1
    Trace("skcache_calc_param: want %lld\n", want);
#endif

#if 1
    Trace("    specified xbk_num %d, xsk_level %d, xffl_num %d\n",
        *xbk_num, *xsk_level, *xffl_num);
#endif

    if(*xffl_size<0) {
        *xffl_size = 1460;
    }

    *xsk_len = (long)
        ( ( want/(*xbk_num) - sizeof(skcache_entry_t)
           - ((sizeof(skcache_frag_t) + *xffl_size) * *xffl_num))
            /
        ((sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(*xsk_level+1)) + sizeof(skcache_entry_t))
        );

#if 1
    Trace("    solved    xffl_size = %d, xsk_len = %d\n",
        *xffl_size, *xsk_len);
#endif


    return *xsk_len;
}

long long
skcache_calc_size(
    int *xbk_num, int *xsk_level, int *xsk_len, int *xffl_len, int *xffl_size)
{
    long long r;
    int esk;
    int tsk;
    int ten;
    int tfl;
    long long ebk;
    long long tbk;

#if 1
    Trace("skcache_calc_size:\n");
    Trace("  bucket:\n");
#endif

    r = (((sizeof(void*)+sizeof(skiplist_cell_t**)+sizeof(skiplist_cell_t*)*(*xsk_level+1)) + sizeof(skcache_entry_t))* (*xsk_len)
        + sizeof(skcache_entry_t)
        + (sizeof(skcache_frag_t) + *xffl_size) * (*xffl_len))*(*xbk_num);

    esk = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(*xsk_level+1);
    tsk = esk * *xsk_len;
#if 1
    Trace("    skiplist:\n");
    Trace("      level %d\n", *xsk_level);
    Trace("      -> element size of skiplist %ld\n", esk);
    Trace("      length of skiplist %d\n", *xsk_len);
#endif

    ten = sizeof(skcache_entry_t)*(*xsk_len+1);
    tfl = (sizeof(skcache_frag_t)+ *xffl_size)*(*xffl_len);
    ebk = (long long)(sizeof(bucket_t)+tsk+ten+tfl);
#if 1
    Trace("    -> size of skiplist            %12ld %4ldM+%07ld %7.2f%%\n",
        tsk, (long)dM(tsk), (long)mM(tsk), ((double)tsk*100)/ebk);
    Trace("    size of free-list for entities %12ld %4ldM+%07ld %7.2f%%\n",
        ten, (long)dM(ten), (long)mM(ten), ((double)ten*100)/ebk);
    Trace("    size of fragment free-list     %12ld %4ldM+%07ld %7.2f%%\n",
        tfl, (long)dM(tfl), (long)mM(tfl), ((double)tfl*100)/ebk);
    Trace("  -> size of bucket                %12lld %4ldM+%07ld\n",
        ebk, (long)dM(ebk), (long)mM(ebk));
#endif


    tbk = ebk * *xbk_num;

#if 1
    Trace("  number of bucket %ld\n", *xbk_num);
    Trace("-> total size of all buckets       %12lld %4ldM+%07ld\n",
        tbk, (long)dM(tbk), (long)mM(tbk));
    Trace("-> total number of entries         %12ld\n", *xsk_len * *xbk_num);
    Trace("-> total number of frags           %12ld\n", *xffl_len * *xbk_num);
#endif

    return r;
}


long long
skcache_init(int xbk_num, int xsk_level, int xsk_len,
        int xffl_num, int xffl_size)
{
    int i;
    int chk;
    long long sum;
    long long all;


#if 1
    Trace("skcache_init:\n\
    bucket: number %d, level %d, length %d\n\
    frag: length %d, size %d\n",
        xbk_num, xsk_level, xsk_len, xffl_num, xffl_size);
#endif


    sum = 0L;

    NIL = (skcache_entry_t*)malloc(sizeof(bucket_t));
    NIL->h1 = INT_MAX;
    NIL->h2 = INT_MAX;
    NIL->h3 = INT_MAX;

    sum += sizeof(bucket_t);

    skcache_bucket_base = (bucket_t*)malloc(sizeof(bucket_t)*xbk_num);
    if(skcache_bucket_base==NULL) {
        Error("not enough memory for skcache_bucket_base");
        skcache_bucket_num = 0;
        return -1;
    }
    skcache_bucket_num = xbk_num;

    sum += sizeof(bucket_t)*xbk_num;

    skcache_entry_freelist_base =
        (skcache_entry_t**)malloc(sizeof(skcache_entry_t)*xbk_num);

    sum += sizeof(skcache_entry_t)*(xbk_num+1);


    skcache_frag_freelist_base =
        (skcache_frag_t**)malloc(sizeof(skcache_frag_t)*xbk_num);
    if(skcache_frag_freelist_base==NULL) {
        Error("not enough memory for skcache_frag_freelist_base");
        return -1;
    }

    sum += sizeof(skcache_frag_t)*xbk_num;

    for(i=0;i<skcache_bucket_num;i++) {
#if 0
Trace("bk# %d %p\n", i, skcache_bucket_base[i]);
#endif

        chk = bk_new(&skcache_bucket_base[i], xsk_level, xsk_len, NIL);
        if(chk==0) {
            Error("initialize error in bucket# %d", i);
            return -2;
        }
        sum += chk;

        chk = alloc_cache_entry_freelist(&skcache_entry_freelist_base[i],
            xsk_len+1);
        if(chk==0) {
            Error("initialize error in object entry freelist# %d", i);
            return -2;
        }
        sum += chk;

        chk = alloc_object_frag_freelist(&skcache_frag_freelist_base[i],
            xffl_num, xffl_size);
        if(chk==0) {
            Error("initialize error in object fragment freelist# %d", i);
            return -2;
        }
        sum += chk;
    }

    Trace("skcache_init: allocate %lld byte(s) %dK %dM %dG\n",
        (long long) sum, (long)dK(sum), (long)dM(sum), (long)dG(sum));

    return sum;
}


off_t
skcache_show_with_HTML(int ofd)
{
    int i;
    off_t wsum, w;
    skiplist_t      *list;
    skiplist_cell_t *cur;
    skcache_entry_t *en;

    wsum = 0;

    w = fdprintf(ofd, "<html>\n");
    wsum += w;



    for(i=0;i<skcache_bucket_num;i++) {
        w = fdprintf(ofd, "<p>bucket# %d\n", i);
        wsum += w;

        w = fdprintf(ofd, "<listing>\n");
        wsum += w;

        list = &skcache_bucket_base[i].cup;
        cur = list->header->skip[0];

        while(cur && cur!= list->term) {
            en = (skcache_entry_t*)cur->entity;
            w = fdprintf(ofd, " %04d:%04d:%04d %d %d '%s'\n",
                    en->h1, en->h2, en->h3, en->ref, en->count, en->name);
            if(w>0) {
                wsum += w;
            }

            cur = cur->skip[0];
        }

        w = fdprintf(ofd, "</listing>\n");
        wsum += w;

    }


    w = fdprintf(ofd, "<html>\n");
    wsum += w;

    Trace("write %ld bytes\n", wsum);
}


