/*
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <malloc.h>
#include <errno.h>

#include "msg.h"
#include "slist.h"

/*
 *
 */


typedef struct _skiplist_cell_t {
    void* entity;
    struct _skiplist_cell_t **skip;
} skiplist_cell_t;

typedef struct _skiplist_t {
    int             maxlevel;
    int             level;
    skiplist_cell_t header;
} skiplist_t;

#define SKL_MAXLEVEL    32





void
dump_skiplist(skiplist_t* list, FILE*fp, char* (fn)(void*) )
{
    skiplist_cell_t* cur;
    int i;
    char label[256];
    char *p;

    if(list==NULL) {
        return;
    }

    fprintf(fp, "\n### list %p, level %d/%d\n",
        list, list->level, list->maxlevel);

    fprintf(fp, "%-24s %-8s ", "header", "header");
    for(i=0;i<=list->level;i++) {
        if(list->header.skip[i]==NULL) {
            fprintf(fp, "%-8s ", "NIL");
        }
        else {
            fprintf(fp, "%p ", list->header.skip[i]);
        }
    }
    fprintf(fp, "\n");

    cur = list->header.skip[0];
    while(cur) {
#if 0
        fprintf(fp, "%p %p ", cur->entity, cur);
#endif
        p = fn(cur->entity);
        if(p) {
            strcpy(label, p);
        }
        else {
            strcpy(label, "*strange*");
        }
        fprintf(fp, "%-24s %p ", label, cur);

        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");

        cur = cur->skip[0];
    }

    fprintf(fp, "%-24s %-8s ", "NIL", "NIL");
    fprintf(fp, "\n\n");
}


skiplist_t*
new_skiplist(int nlevel)
{
    skiplist_t* this;
    int i;

    this = (skiplist_t*) malloc(sizeof(skiplist_t)*1);
    if(this==NULL) {
        return NULL;
    }
    if(nlevel>SKL_MAXLEVEL) {
        nlevel = SKL_MAXLEVEL;
    }

    this->maxlevel  = nlevel;
    this->level     = 0;
    this->header.entity = NULL;
    this->header.skip =
            (skiplist_cell_t**) malloc(sizeof(skiplist_cell_t*)*nlevel);
    if(this->header.skip==NULL) {
        free(this);
        return NULL;
    }
    for(i=0;i<this->maxlevel;i++) {
        this->header.skip[i] = (skiplist_cell_t*)NULL;
    }

    return this;
}


static
int
randomlevel(int max, double p)
{
    int lv;
    int w = 1000;
    int v;

    lv = 0;
    while( v=rand()%w, v<(int)((double)w*p) && lv < max) {
        lv++;
    }

    return lv;
}


/*
 * cmp - compare
 * rep - replace
 */
int
insert_skiplist(skiplist_t *list, void* newface,
        int eqflag,
        int (cmp)(void*,void*), int (rep)(void*,void*))
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x;
    int i;
    int newlevel;

    /* find insert point and update points at each level */
    for(i=0;i<list->maxlevel;i++) {
        update[i] = NULL;
    }
    x = &list->header;
    for(i=list->level; i>=0; i--) {
        while(x->skip[i] && cmp(x->skip[i]->entity, newface)<0) {
            x = x->skip[i];
        }
        update[i] = x;
    }
    x = x->skip[0];


    if(x && !eqflag && cmp(x->entity, newface)==0) {
        rep(x->entity, newface);
    }
    else {

        newlevel = randomlevel(list->maxlevel, 0.5);

        if(newlevel > list->level) {
            for(i=list->level+1;i<=newlevel;i++) {
                update[i] = &list->header;
            }
            list->level = newlevel;
        }


        x = (skiplist_cell_t*)malloc(sizeof(skiplist_cell_t));
        if(x==NULL) {
            return -1;
        }
        x->entity = newface;
        x->skip =
            (skiplist_cell_t**)malloc(sizeof(skiplist_cell_t*)*list->maxlevel);
        if(x->skip==NULL) {
            free(x);
            return -1;
        }

        for(i=0;i<=newlevel;i++) {
            x->skip[i] = update[i]->skip[i];
            update[i]->skip[i] = x;
        }
    }

    return 0;
}


/*
 * cmp - compare
 * rem - remove
 */
int
delete_skiplist(skiplist_t *list, void* newface,
        int (cmp)(void*,void*), int (rem)(void*))
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x;
    int i;
    int newlevel;

    /* find insert point and update points at each level */
    for(i=0;i<list->maxlevel;i++) {
        update[i] = NULL;
    }
    x = &list->header;
    for(i=list->level; i>=0; i--) {
        while(x->skip[i] && cmp(x->skip[i]->entity, newface)<0) {
            x = x->skip[i];
        }
        update[i] = x;
    }
    x = x->skip[0];


    if(x && cmp(x->entity, newface)==0) {
        for(i=0;i<list->level;i++) {
            if(update[i]->skip[i]!=x) {
                break;
            }
            else {
                update[i]->skip[i] = x->skip[i];
            }
        }
        rem(x);
        while(list->level>0 && list->header.skip[list->level]==NULL) {
            list->level--;
        }
    }

    return 0;
}


#if 1

typedef struct {
    char key[16];
    int val;
} dummy_t;

#define KOTETU_SUB_TIME(x,y) \
        ( (x.tv_sec  - y.tv_sec )*1000 + (x.tv_usec - y.tv_usec)/1000 )


char *dummy_label(void* x)
{
    static char buf[256];
    dummy_t *pos;

    pos = (dummy_t*) x;

    sprintf(buf, "'%s'", pos->key);

    return buf;
}


int dummy_cmp(void *px, void *py)
{
    dummy_t *x, *y;
    int a, b;
    int r;

    x = (dummy_t*)px;
    y = (dummy_t*)py;

    if(x==NULL) {
        r = -1;
    }
    else
    if(y==NULL) {
        r = 1;
    }
    else {
#if 0
        r = strcmp(x->key, y->key);
#else
        a = atoi(x->key);
        b = atoi(y->key);
        r = a - b;
#endif
    }

    return r;
}

int dummy_replace(void *px, void *py)
{
    dummy_t *x, *y;

    x = (dummy_t*)px;
    y = (dummy_t*)py;

    *x = *y;
}

int dummy_free(void *px)
{
    dummy_t *x;

    x = (dummy_t*)px;
}

void
dummy_show(skiplist_t *list, FILE *fp)
{
    skiplist_cell_t* cur;
    int i;
    char label[256];
    char *p;

    if(list==NULL) {
        return;
    }

    cur = list->header.skip[0];
    while(cur) {
        fprintf(fp, "%4d\n", ((dummy_t*)cur->entity)->val);

        cur = cur->skip[0];
    }
}

main()
{
    skiplist_t *list;
    char line[1024];
    dummy_t *index[100000];
    dummy_t *foo;
    int chk;
    int c;
    int i;
    struct timeval begin, end;
    FILE *fp;

    list = new_skiplist(30);
    if(list==NULL) {
        fprintf(stdout, "cannot get new list\n");
        exit(1);
    }

#if 0
    gettimeofday(&begin, NULL);
#endif

    fp = fopen("x", "r");
    if(fp==NULL) {
        exit(3);
    }

    c = 0;
    while(fgets(line, 1024, fp)!=NULL) {
        foo = (dummy_t*)malloc(sizeof(dummy_t));
/*
        strcpy(foo->key, line);
*/
        foo->val = atoi(line);
        sprintf(foo->key, "%d", foo->val);
#if 0
        fprintf(stdout, "=== %4d: %9d ===\n", c, foo->val);
#endif
        chk = insert_skiplist(list, foo, 1, dummy_cmp, dummy_replace);
        if(chk<0) {
            fprintf(stdout, "cannot insert new entity\n");
        }
#if 0
        dump_skiplist(list, stdout, dummy_label);
#endif
        c++;
    }

    fclose(fp);

    dummy_show(list, stdout);
#if 0
    fprintf(stdout, "# %d items\n", c);
#endif


#if 0
    gettimeofday(&end, NULL);

    fprintf(stderr, "takes %d sec\n", end.tv_sec - begin.tv_sec);
    fprintf(stderr, "takes %d msec\n", KOTETU_SUB_TIME(end, begin));
#endif

}

#endif
