/*
 * xsk.c --- sk.c + skcache.c
 */

/*
 * sk.c --- skip list procedures
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 */

/*
 * skcache.c --- object cache procedures using skip list and its bucket
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 *
 * $Id$
 */

/*
 * See also:
 *   William Pugh, "skiplist cookbook",
 *   UMIACS-TR-89-72.1, July 1989
 */

/*
 * The algorithm on the original text was described by Pascal like
 * langauage.  Subscription were 1..maxlevel or 1..level .
 * Since this is C program,  This file describes subscription as
 * 0..maxlevel-1 and 0..level-1,
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <dirent.h>

#include <malloc.h>
#include <assert.h>
#include <errno.h>

#include "range.h"
#include "msg.h"

#include "xsk.h"
#include "sim_stat.h"

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif
#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

#ifndef DISK_RETRY
#define DISK_RETRY (13)
#endif


/*
 * arround 256, follwing numbers are known as prime numbers.
 *  241 251 257 263 269 271
 *
 */

int radix_h1 =   257;
int radix_h2 =   263;
int radix_h3 =   251;
int radix_h4 =   499;

skcache_entry_t *NIL;
int skcache_bucket_num;
bucket_t *skcache_bucket_base;

extern pthread_mutex_t stat_lock;

volatile int swapload_count     = 0;
volatile int swapload_success   = 0;
volatile int swapin_count       = 0;
volatile int swapin_success     = 0;
volatile int swapout_count      = 0;
volatile int swapout_success    = 0;

sim_stat_rec    swapload_time;
static long     swapload_time_count[K(1)];
sim_stat_rec    swapin_time;
static long     swapin_time_count[K(1)];
sim_stat_rec    swapout_time;
static long     swapout_time_count[K(1)];
sim_stat_rec    compaction_time;
static long     compaction_time_count[K(2)];
sim_stat_rec    checkdir_time;
static long     checkdir_time_count[K(2)];
sim_stat_rec    checkmem_time;
static long     checkmem_time_count[K(2)];


int require_radix_h2;
int require_radix_h3;
off_t require_dir_size;
int require_swap_token_val;
int require_swap_token_interval;
int require_disk_sweep_interval;



extern pthread_mutex_t config_lock;

int skcache_mem_keep_min    = 20;
/*
int skcache_mem_keep_cur    = 240;
*/
int skcache_mem_keep_cur    = 1200;
int skcache_mem_keep_max    = 3600;
int skcache_entry_keep_min  = 120;
int skcache_entry_keep_cur  = 36000;
int skcache_entry_keep_max  = 86400;

volatile long long swap_count = 0LL;
volatile long long swap_size = 0LL;

/* limitter */
volatile int swapout_token_init = 1000;
volatile int swapout_token	= 1000;
volatile int swapout_token_interval;
volatile int disk_sweep_interval;

time_t min_date;

int *dir_sig;
int *dir_sz;
int dir_num=0;

void
sk_rawdump(FILE* fp, skiplist_t *list, char* (fn)(void*))
{
    skiplist_cell_t *cur;
    int c, i;

#if 0
    for(i=0;i<list->level;i++) {
        fprintf(stdout, " h %4d: %p\n", i, list->header->skip[i]);
    }
#endif

    fprintf(fp, "--- list addr %p, num %d, func %p\n",
        list, list->num, fn);

    c = 0;
    cur = list->header->skip[0];
    if(fn)
        while(cur != list->term) {
            fprintf(fp, " %4d: %p : %p %s // %p\n",
                c, cur, cur->entity, fn(cur->entity), cur->skip[0]);
            cur = cur->skip[0];
            c++;
        }
    else
        while(cur != list->term) {
            fprintf(fp, " %4d: %p : %p // %p\n",
                c, cur, cur->entity, cur->skip[0]);
            cur = cur->skip[0];
            c++;
        }
}


void
sk_dump(skiplist_t* list, FILE*fp, char* (fn)(void*) )
{
    skiplist_cell_t* cur;
    int i;
    char label[256];
    char *p;
    int k;

    if(list==NULL) {
        return;
    }

    fprintf(fp, "### list %p, num %d, level %d/%d\n",
        list, list->num, list->level, list->maxlevel);

    fprintf(fp, "%-24s %-10s ", "header", "header");
    for(i=0;i<=list->level;i++) {
        if(list->header->skip[i]==NULL) {
            fprintf(fp, "%-10s ", "NIL");
        }
        else {
            fprintf(fp, "0x%08x ", list->header->skip[i]);
        }
    }
    fprintf(fp, "\n");

#if 1
    cur = list->header->skip[0];
    while(cur) {
#if 0
        fprintf(fp, "%p %p ", cur->entity, cur);
#endif
        if(fn!=NULL) {
            p = fn(cur->entity);
            if(p) {
                strcpy(label, p);
            }
            else {
                strcpy(label, "*strange*");
            }
        }
        else {
            strcpy(label, "*NONE*");
        }


        fprintf(fp, "%-24s 0x%08x ", label, cur);

        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-10s ", "NIL");
            }
            else {
                fprintf(fp, "0x%08x ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");

		if(cur==list->term) {
			break;
		}
		if(cur==cur->skip[0]) {
			fprintf(fp, "*** looped\n");
			break;
		}

        cur = cur->skip[0];
    }

    fprintf(fp, "%-24s %-10s ", "NIL", "NIL");
    fprintf(fp, "\n\n");
#endif

}

void
bk_dump(bucket_t *bk, skiplist_t* list, FILE*fp, char* (fn)(void*) )
{
    char *pcur;
    skiplist_cell_t* cur;
    int i;
    char label[256];
    char *p;
    int j, k;

    if(list==NULL) {
        return;
    }

    fprintf(fp, "### list %p, num %d, level %d/%d, term %p\n",
        list, list->num, list->level, list->maxlevel, list->term);

#ifndef DEBUG
    return;
#endif

#ifdef DEBUG
    fprintf(fp, "#  first %p last %pd\n#\n",
        bk->cell_first, bk->cell_last);

#if 0
    if(fn==NULL) {
        fprintf(fp, "#%6s %8s %8s %s %p\n",
            "", "", "", "header", list->header);
        fprintf(fp, "#%6s %8s %8s ",
            "", "", "");
    }
    else {
        fprintf(fp, "#%6s %8s %8s %-14s %s %p\n",
            "", "", "", "", "header", list->header);
        fprintf(fp, "#%6s %8s %8s %-14s ",
            "", "", "", "");
    }
    if(list->level<0) {
        fprintf(fp,"*** no level ***\n");
    }
    else {

        cur = list->header;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
    fprintf(fp, "\n");

    }
#endif

#if 1
    fprintf(fp, "# cup  header %p: ", bk->cup.header);
        cur = bk->cup.header;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");
    fprintf(fp, "# cup  term   %p: ", bk->cup.term);
        cur = bk->cup.term;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");

#endif


    fprintf(fp, "#\n");

    if(fn==NULL) {
        fprintf(fp, "#%6s %8s %8s %8s",
            "num", "addr", "entity", "skips..");
    }
    else {
        fprintf(fp, "#%6s %8s %8s %-14s %8s",
            "num", "addr", "entity", "label", "skips..");
    }
    fprintf(fp, "\n");

    j = -1;
    k = 0;
    pcur = (char*)bk->cell_first;
    while(pcur<=(char*)bk->cell_last) {
        cur=(skiplist_cell_t*)pcur;

        j++;
        if(cur->entity==NULL) {
            goto next;
        }
#if 0
        fprintf(fp, "%p %p ", cur->entity, cur);
#endif
        if(fn!=NULL) {
            p = fn(cur->entity);
            if(p) {
                strcpy(label, p);
            }
            else {
                strcpy(label, "*strange*");
            }
            fprintf(fp, "= %6d %6d %p %p %-14s ",
                j, k, cur, cur->entity, label);
        }
        else {
            fprintf(fp, "= %6d %6d %p %p ",
                j, k, cur, cur->entity);
        }

        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "x");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");
        k++;
#if 0
        if(k>list->num+3) {
            break;
        }
#endif

next:
        pcur += bk->elemsize;
    }

    fprintf(fp, "\n\n");
    fflush(fp);

#endif
}


int
check_bkentry(bucket_t *bk, char *prefix)
{
	int i, rest, diff;
	skcache_entry_t *cur;

	if(bk==NULL) {
		return 1;
	}

	cur = bk->entry_freelist;
	i = 0;
	while(cur) {
		if(cur==cur->next) {
			Error(" %d th looped", i);
		}
		i++;
		cur = cur->next;
	}
	rest = i;

	diff = bk->entry_all - (bk->entry_use+rest);

	Trace("%s bk# %d, %p, rest %d, %d/%d/%d -> %d, diff %d\n",
		prefix,
		bk->bid, bk, rest,
		bk->entry_use, bk->entry_hwater, bk->entry_all,
		bk->entry_use+rest, diff);

	if(diff==0)
		return 1;
	else 
		return 0;
}


int
dump_bkentryfreelist(bucket_t *bk, char * prefix)
{
	int i, rest, diff;
	skcache_entry_t *cur;

	cur = bk->entry_freelist;
	i = 0;
	while(cur) {
		Trace("%4d %08x %08x  ", i, cur, cur->next);
		
		if(cur==cur->next) {
			Error(" %d th looped", i);
		}
		i++;
		cur = cur->next;
	}
	Trace("\n");
	rest = i;

	diff = bk->entry_all - (bk->entry_use+rest);

	Trace("%s bk# %d, %p, rest %d, %d/%d/%d -> %d, diff %d\n",
		prefix,
		bk->bid, bk, rest,
		bk->entry_use, bk->entry_hwater, bk->entry_all,
		bk->entry_use+rest, diff);

	if(diff==0)
		return 1;
	else 
		return 0;
}

static
int
randomlevel(int max, double p)
{
    int lv;
    int w = 4097;
    double v;

    lv = 0;
#if 0
    while( v=((double)(rand()%w))/w, v < p && lv < max) {
        lv++;
    }
#endif
    while(drand48() < p && lv < max) {
        lv++;
    }

    return lv;
}




/*
 * cmp - compare
 */
skiplist_cell_t *
sk_find_listpos(skiplist_t *list, void* newface,
        int (cmp)(void*,void*))
{
    register skiplist_cell_t *x;
    register int i, diff;
    skiplist_cell_t *ret;

    ret = NULL;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        while(cmp(x->skip[i]->entity, newface)<0) {
            x = x->skip[i];
        }
        /* XXX */
        if(diff==0) {
            ret = x->skip[i];
            goto out;
        }
    }
    x = x->skip[0];

    if(cmp(x->entity, newface)==0) {
        ret = x;
    }

out:
    return ret;
}

void*
sk_find(skiplist_t *list, void* newface,
        int (cmp)(void*,void*))
{
    register skiplist_cell_t *x;
    register int i, diff;
    void *ret;

    ret = NULL;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        while(diff = cmp(x->skip[i]->entity, newface), diff<0) {
            x = x->skip[i];
        }
        /* XXX */
        if(diff==0) {
            ret = x->skip[i]->entity;
            goto out;
        }
    }
    x = x->skip[0];

    if(cmp(x->entity, newface)==0) {
        ret = x->entity;
    }

out:
    return ret;
}

/*
 * if alreay assigned, not new allocation.
 * cmp - compare
 */
void*
sk_assign(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*),
        int* st, int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    register skiplist_cell_t *x;
    register int i;
    int newlevel;
    skiplist_cell_t *nextfree;
    skiplist_cell_t *ret;
    int c;
    int diff;

#if 1
	if(newface==NULL) {
		Error("don't assign NULL to skiplist\n");

		ret = NULL;
		*st = -1;
		goto out;
	}
#endif

	if(*flist==NULL) {
		fprintf(stderr, "no free cell (opt %d)\n", opt);
		ret = NULL;
		*st = -1;
		goto out;
	}

    x = list->header;
    for(i=list->level; i>=0; i--) {
        c = 0;

        while(1) {
#if 0
            diff = cmp(x->skip[i]->entity, newface);
#endif
			if(x->skip[i]->entity) {
           		diff = cmp(x->skip[i]->entity, newface);
			}
			else {
				diff = INT_MAX;
			}

            if(diff>=0) {
                break;
            }
            x = x->skip[i];
            c++;
        }
        update[i] = x;
    }
    x = x->skip[0];

assert(x && x->entity);

    if(cmp(x->entity, newface)==0) {
        ret = x->entity;
        *st = 1;
    }
    else {
        newlevel = randomlevel(list->maxlevel, 0.5);

assert(newlevel<=list->maxlevel);

        if(newlevel > list->level) {
            for(i=list->level+1;i<=newlevel;i++) {
                update[i] = list->header;
            }
            list->level = newlevel;
        }

        if(*flist==NULL) {
            fprintf(stderr, "no free cell (opt %d)\n", opt);
            ret = NULL;
            *st = -1;
            goto out;
        }

        nextfree = (*flist)->skip[0];

        x = *flist;
        *flist = nextfree;

        x->entity = newface;
        list->num++;

        ret = x->entity;
        *st = 0;

        for(i=0;i<=newlevel;i++) {
            x->skip[i] = update[i]->skip[i];
            update[i]->skip[i] = x;
        }

    }

out:
    return ret;
}

int
sk_insert(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*), int (ovw)(void**,void*), int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    register skiplist_cell_t *x;
    register int i;
    int newlevel;
    skiplist_cell_t *nextfree;
    int c;
    int diff;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        c = 0;

        while(1) {
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
            c++;
        }
        update[i] = x;
    }
    x = x->skip[0];

assert(x && x->entity);

    if(cmp(x->entity, newface)==0 && ovw) {
        ovw(&x->entity, newface);
    }
    else {
        newlevel = randomlevel(list->maxlevel, 0.5);

assert(newlevel<=list->maxlevel);

        if(newlevel > list->level) {
            for(i=list->level+1;i<=newlevel;i++) {
                update[i] = list->header;
            }
            list->level = newlevel;
        }

        if(*flist==NULL) {
            fprintf(stderr, "no free cell (opt %d)\n", opt);
            return -1;
        }

        nextfree = (*flist)->skip[0];

        x = *flist;
        *flist = nextfree;

        x->entity = newface;
        list->num++;

        for(i=0;i<=newlevel;i++) {
            x->skip[i] = update[i]->skip[i];
            update[i]->skip[i] = x;
        }

    }

    return 0;
}

/*
 * cmp - compare
 * rem - remove
 */
void*
sk_delete(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*), int (rem)(void*), int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x;
#if 0
    skiplist_cell_t *y;
#endif
    int i;
    void *ret;
    int diff;
    int chk;

    ret = NULL;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        while(diff = cmp(x->skip[i]->entity, newface), diff<0) {
            x = x->skip[i];
            if(x->skip[i]==NULL)
                break;
        }
        update[i] = x;
    }
    x = x->skip[0];

    if(!x) {
        return ret;
    }

    if(cmp(x->entity, newface)==0) {
        if(rem) {
            chk = rem(x->entity);
            if(chk) {
                return ret;
            }
        }

        for(i=0;i<=list->level;i++) {
            if(update[i]->skip[i]==x) {
                update[i]->skip[i] = x->skip[i];
            }
        }

        while(list->level>0 &&
                list->header->skip[list->level]==list->term) {
            list->level--;
        }

        ret = x->entity;
        x->entity = NULL;
        for(i=1;i<=list->maxlevel;i++) {
            x->skip[i] = NULL;
        }
        x->skip[0] = *flist;
        *flist = x;

        list->num--;
    }

    return ret;
}

/*
 * ack - decide delete or not
 *          that function have to return not zero for delete
 */
int
sk_mapdelete(skiplist_t *list,
        skiplist_cell_t **flist,
        int (ack)(void*, void*), void *opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x, *y, *n;
    int i;
    int ret;
    int diff;
    int chk;

    ret = 0;

    x = list->header;
    for(i=0;i<=list->level;i++) {
            update[i] = x;
    }

    x = x->skip[0];

    while(x!=list->term) {
        n = x->skip[0]; /* save next element */

        if(ack==NULL)
            goto delete;

		chk = ack(x->entity, opt);

        if(chk) {
delete:
            /*
             * skip current element
             * so, it is mean deletation of element
             * this routine don't care actual enitity procedure
             * (e.g, free(), unlink() or something )
             */
            for(i=0;i<=list->level;i++) {
                if(update[i]->skip[i]==x) {
                    update[i]->skip[i] = x->skip[i];
                }
            }

            /* add current element to free-list for recycle */
            y = x;
            y->entity = NULL;
            for(i=1;i<=list->maxlevel;i++) {
                y->skip[i] = NULL;
            }
            y->skip[0] = *flist;
            *flist = y;

            list->num--;

            ret++;
        }
no_delete:

        for(i=0;i<=list->level;i++) {
            if(update[i]->skip[i]==x)
                update[i] = x;
        }

        x = n;
    }

    return ret;
}


int
Zsk_mapdelete(skiplist_t *list,
        skiplist_cell_t **flist,
        int (ack)(void*, void*), void *opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x, *y, *n;
    skiplist_cell_t *last;
    int i;
    int ret;
    int diff;
    int chk;
	int c;

char* skcache_label4_ref(void *pp);


    ret = 0;

#if 0
	Trace("\n\n,,,\n");
	sk_dump(list, stderr, skcache_label4_ref);
#endif

	list->level = 0;
    x = list->header;
	x->entity = list->term;
    for(i=1;i<=list->maxlevel;i++) {
		x->skip[i] = list->term;
    }

	last = NULL;
    x = x->skip[0];
	c = 0;

    while(x!=list->term) {
		c++;
        n = x->skip[0]; /* save next element */

/*
Trace("- %4d: %p n %p l %p\n", c, x, x->skip[0], last);
*/

        for(i=1;i<=list->maxlevel;i++) {
			x->skip[i] = list->term;
		}

		if(x->entity==NULL) {
Error("list %p, found NULL in %d th", list, c);
			chk = 1;
		}
		else {
        	chk = ack(x->entity, opt);
		}

        if(chk) {
/*
Trace("  *delete*\n");
*/
			if(last==NULL) {
				list->header->skip[0] = n;
			}
			else {
				last->skip[0] = n;
			}

			x->entity  = NULL;
            x->skip[0] = *flist;
            *flist = x;

            list->num--;

            ret++;
        }
		else {
			last = x;
		}

        x = n;
    }

#if 0
	sk_dump(list, stderr, skcache_label4_ref);

	Trace("return %d\n", ret);
#endif
/*
abort();
*/

    return ret;
}


int
Xsk_mapdelete(skiplist_t *list,
        skiplist_cell_t **flist,
        int (ack)(void*), int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x, *y;
    int i;
    int ret;
    int diff;
    int chk;

    ret = 0;

    x = list->header;
    for(i=0;i<=list->level;i++) {
            update[i] = x;
    }

    do {
        for(i=0;i<=list->level;i++) {
            if(update[i]->skip[i]==x)
                update[i] = x;
        }

        x = x->skip[0];
        if(x==list->term)
            break;

        chk = ack(x->entity);
fprintf(stderr, "x %p, entity %p, chk %d\n", x, x->entity, chk);

fprintf(stderr, "\t");
 for(i=0;i<=list->level;i++) {
    fprintf(stderr, "%p ", update[i]);
 }
fprintf(stderr, "\n");


        if(chk) {

            for(i=0;i<=list->level;i++) {
                if(update[i]->skip[i]==x) {
                    update[i]->skip[i] = x->skip[i];
                }
            }



            ret++;
        }
    } while(x!=list->term);

    return ret;
}

skiplist_t*
sk_new(int nlevel, void* user_term)
{
    skiplist_t *this;
    skiplist_cell_t *n;
    int i, j;

    this = (skiplist_t*) malloc(sizeof(skiplist_t)*1);
    if(this==NULL) {
        return NULL;
    }
    if(nlevel>SKL_MAXLEVEL-1) {
        nlevel = SKL_MAXLEVEL-1;
    }

    this->maxlevel  = nlevel;
    this->level     = 0;
    n = this->header;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)malloc(
                    sizeof(void*)+sizeof(skiplist_cell_t**)
                    +sizeof(skiplist_cell_t*)*(nlevel+1));
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = this->term;
        }
    n = this->term;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)malloc(
                    sizeof(void*)+sizeof(skiplist_cell_t**)
                    +sizeof(skiplist_cell_t*)*(nlevel+1));
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = this->term;
        }

    return this;
}


int
bucket_show(bucket_t *bk)
{
    skiplist_cell_t *n;
    skiplist_cell_t **p;
    int i, j;

    fprintf(stdout, "=== bk %p\n", bk);

    fprintf(stdout, "cup  header,skip: %p\n", bk->cup.header->skip);
    n  = bk->cup.header->skip[0];
    i  = 0;
    while(n) {
        fprintf(stdout, "%4d: <%p> %p %p\n",
            i, n, n->entity, n->skip[0]);
        n = n->skip[0];
        i++;
    }
    fprintf(stdout,"  have %d item(s)\n", i);

    fprintf(stdout, "free list: %p\n", bk->elem_freelist);

    n  = bk->elem_freelist;
    i  = 0;
    while(n) {
        fprintf(stdout, "%4d: <%p> %p %p\n",
            i, n, n->entity, n->skip[0]);
        n = n->skip[0];
        i++;
    }
    fprintf(stdout,"  have %d item(s)\n", i);

    return 0;
}

int
bk_new(bucket_t *bk, int nlevel, int nelement, void* user_term)
{
    skiplist_cell_t *n;
    skiplist_cell_t **p;
    int i, j;
    char *pool;


    if(bk==NULL) {
        return 0;
    }

    if(nlevel>SKL_MAXLEVEL-1) {
        nlevel = SKL_MAXLEVEL-1;
    }

    bk->elemnum = nelement;
    bk->elemsize = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(nlevel+1);
    bk->totalsize = bk->elemsize*(nelement+4);
    pool = (char*)malloc(bk->totalsize);
    if(pool==NULL) {
        fprintf(stderr, "not enough memory !!!\n");
        return -1;
    }

#if 0
    fprintf(stderr, "nlevel %d, nelement %d\n",
            nlevel, nelement);
    fprintf(stderr, "element size %d, num %d -> total %d\n",
            bk->elemsize, nelement, bk->totalsize);
    fprintf(stderr, "pool  [%p..%p)\n", pool, pool+bk->totalsize);
#endif
#if 0
    for(i=0;i<nelement+2;i++) {
        fprintf(stdout, "%4d: <%p>\n", i, pool+bk->elemsize*i);
    }
    fflush(stdout);
#endif
    if(pthread_mutex_init(&bk->mutex, NULL)<0) {
        Error("fail pthread_mutex_init() (%d)", errno);
    }

    bk->cup.num         = 0;
    bk->cup.maxlevel    = nlevel;
    bk->cup.level       = 0;
    bk->cup.header      = (skiplist_cell_t*)pool;
    bk->cup.term        = (skiplist_cell_t*)(pool + 1*bk->elemsize);
    n = bk->cup.header;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = bk->cup.term;
        }
    n = bk->cup.term;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = bk->cup.term;
        }


#if 0
    fprintf(stderr, "cup  header %p\n", bk->cup.header);
    fprintf(stderr, "cup  term   %p\n", bk->cup.term);
#endif

    bk->elem_freelist = (skiplist_cell_t*)(pool+4*bk->elemsize);
#if 0
    fprintf(stderr, "elem_freelist    %p\n", bk->elem_freelist);
#endif


#ifdef DEBUG
    bk->cell_first = (skiplist_cell_t*)(pool+0*bk->elemsize);
    bk->cell_last  = (skiplist_cell_t*)(pool+nelement*bk->elemsize);

    fprintf(stderr, "  first..last %p..%p\n", bk->cell_first, bk->cell_last);
#endif

    n = bk->elem_freelist;
    for(i=0;i<nelement;i++) {
        if(n==NULL) {
            break;
        }
#if 0
        fprintf(stdout, "* %d, %p\n", i, n);
        fflush(stdout);
#endif
        n->entity = NULL;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=1;j<nlevel;j++) {
            n->skip[j] = NULL;
        }
        if(i<nelement-1) {
            n->skip[0] = (skiplist_cell_t*)((char*)n+bk->elemsize);
            n = (skiplist_cell_t*)((char*)n+bk->elemsize);
        }
        else {
            n->skip[0]  = NULL;
        }
    }

#if 0
    fprintf(stdout, "cup  header,skip: %p\n", bk->cup.header->skip);
    fprintf(stdout, "free: %p\n", bk->elem_freelist);

    n  = bk->elem_freelist;
    for(i=0;i<nelement;i++) {
        if(n==NULL)
            break;

#if 0
        fprintf(stdout, "%4d: <%p> ", i, n);
        for(j=0;j<nlevel;j++) {
            fprintf(stdout, "%d:<%p> %p ",
                j, &(n->skip[j]), n->skip[j]);
        }
        fprintf(stdout, "\n");
#endif
#if 1
        fprintf(stdout, "%4d: <%p> %p\n", i, n, n->skip[0]);
#endif

        n = n->skip[0];
    }

#endif

    return bk->totalsize;
}


/*
{23,19,13,31,11,67,7,71,5,3,29};
{7,191,17,13,27,127,83,241,59,281,293,19,};
{229,113,53,181,257,149,137,293,223,263,211,};
{22,113,251,181,7,149,137,5,223,263,211,};
*/


/*
{ 181, 157, 7, 19, 71, 229, 167, 41, 61, 251, 283,};
{ 43, 281, 109, 5, 5, 71, 199, 293, 193, 83, 157,};
{ 197, 107, 251, 131, 11, 3, 227, 263, 293, 281, 223,};
{ 149, 271, 73, 137, 251, 97, 19, 5, 263, 73, 2,};
{ 79, 73, 2, 179, 3, 293, 79, 97, 151, 5, 229,};
{ 97, 139, 239, 101, 67, 193, 83, 53, 109, 271, 23,};
*/


/* used for index */
/*
static int prime_box[]  ={ 149,271,73,137,251,97,19,5,263,73,2,};
static int prime_box2[] ={ 79,73,2,179,3,293,79,97,151,5,229,};
static int prime_box3[] ={ 197,107,251,131,11,3,227,263,293,281,223,};
static int prime_box4[] ={ 181,157,7,19,71,229,167,41,61,251,283,};
*/

/*
	{2025,14156,3428,613,14090,6233,1998,4068,2745,12918,};
	{5768,4715,4463,7697,1123,3127,14391,7537,6446,7292,};
	{4937,3480,4233,1123,10671,8817,1637,100,13957,210,};
	{13450,4758,8665,3930,6864,529,7150,772,8123,12810,};
*/

/*
	{983,2531,71,1609,8641,907,9371,11731,521,4211,};
	{9413,2377,1949,4787,647,6353,12251,10979,439,383,};
	{3373,9677,2081,31,7013,5881,7207,9203,4801,7741,};
	{9239,12203,431,7351,12113,3307,991,9749,4931,11633,};
*/

static int prime_box[]  =
    {162059,74941,53479,117431,110581,114083,92431,88463,73,152623,};
static int prime_box2[] =
    {81611,6287,55381,46027,30949,108061,32573,66739,55547,83389,};
static int prime_box3[] =
    {25087,124769,62297,123377,71453,144497,88397,44257,119183,11317,};
static int prime_box4[] =
    {55967,8287,71341,129293,34259,7607,84191,4721,75557,34259,};

int
calc_hash(unsigned char *b, int *h1, int *h2, int *h3, int *h4)
{
    int l;
    register int sum, c;
    register unsigned char *p;
    register unsigned char q;

    sum = 0;

    c = 1;
    p = b;
    q = 127;
    while(*p) {
#if 1
        sum += *p - ' ' + c;
        if(*p>='0' && *p<='9') {
            sum += prime_box[(c+*p)%10] + prime_box2[*p - '0'];
        }
		else {
	    	sum += - prime_box3[*p%10] - prime_box4[c%10];
		}
#endif
#if 0
        sum += sum % prime_box2[(c+*p)%10];
        sum += q + *p + prime_box4[*p%10] + prime_box3[c%10];
#endif
        q = *p;

        c++;
        p++;
    }

    if(sum<0)
        sum = -sum;
	
	*h1 = sum%radix_h1;
	*h2 = sum%radix_h2;
	*h3 = sum%radix_h3;
	*h4 = sum%radix_h4;

#if 0
Trace("sum %9d -> %3d, %3d, %3d, %3d\n", sum, *h1, *h2, *h3, *h4);
#endif

    return sum % radix_h1;
}


int
first_hash(unsigned char *b)
{
    int l;
    register int sum, c;
    register unsigned char *p, *r, *s;

    sum = 0;

    c = 1;
    p = b;
    while(*p) {
        sum += *p - ' ' + c;
        if(*p>='0' && *p<='9') {
            sum += prime_box2[*p - '0'];
        }
		else {
			sum += prime_box3[*p%10];
		}
        p++;
        c++;
    }

    if(sum<0)
        sum = -sum;

    return sum % radix_h1;
}

int
second_hash(unsigned char *b)
{
    register int sum;
    register unsigned char *p;
    register int c;

    sum = 0;
    c = 1;

    p = b;
    while(*p) {
        if(*p>='0' && *p<='9') {
        	sum += prime_box3[*p - '0'];
		}
		else {
			sum += prime_box4[*p%10];
		}
        c++;
        p++;
    }

    if(sum<0)
        sum = -sum;

    return sum % radix_h2;
}

int
third_hash(unsigned char *b)
{
    register int sum;
    register unsigned char *p;
    register int c;

    sum = 0;
    c = 1;

    p = b;
    while(*p) {
		sum = sum << 2 ;
        sum += *p + c;

        c++;
        p++;
    }

    if(sum<0)
        sum = -sum;

    return sum % radix_h3;
}



int 
fourth_hash(unsigned char *b)
{   
    register int sum;
    register unsigned char *p;
    register unsigned char q;
    register int c;

    sum = 0;
    c = 1;

    p = b;
    q = 127;
    while(*p) {
        sum += sum % prime_box2[(c+*p)%10];
        sum += q + *p + prime_box4[*p%10] + prime_box3[c%10];
        q = *p;
        c++;
        p++;
    }

    if(sum<0) 
        sum = -sum;

    return sum % radix_h4;
}








int
skcache_entry_cmp(void *px, void *py)
{
    skcache_entry_t *x = (skcache_entry_t*)px;
    skcache_entry_t *y = (skcache_entry_t*)py;

    if(x->h1==y->h1) {
        if(x->h2==y->h2) {
            if(x->h3==y->h3) {
                return strcmp((const char*)x->name, (const char*)y->name);
            }
            else {
                return x->h3 - y->h3;
            }
        }
        else {
            return x->h2 - y->h2;
        }
    }
    else {
        return x->h1 - y->h1;
    }
}

int
X_skcache_entry_cmp(void *px, void *py)
{
    skcache_entry_t *x = (skcache_entry_t*)px;
    skcache_entry_t *y = (skcache_entry_t*)py;

    if(x->h1==y->h1) {
        if(x->h2==y->h2) {
            return x->h3 - y->h3;
        }
        else {
            return x->h2 - y->h2;
        }
    }
    else {
        return x->h1 - y->h1;
    }
}

char*
skcache_label_hash(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_PATH];

    sprintf(tmp, "%04d:%04d:%04d", p->h1, p->h2, p->h3);

    return tmp;
}

char*
skcache_label_short(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d %s", p->h1, p->h2, p->h3, p->name);

    return tmp;
}

char*
skcache_label_ref(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d-%03d", p->h1, p->h2, p->h3, p->ref);

    return tmp;
}

char*
skcache_label4_ref(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d:%04d-%03d",
		p->h1, p->h2, p->h3, p->h4, p->ref);

    return tmp;
}


int
skcache_dump_entry_frag(char *msg, skcache_entry_t *en)
{
    skcache_frag_t *p;
    off_t frag_sum;
    int u;

    Trace("%s\n", msg);
    Trace("  frags of entry %p\n", en);
    if(en==NULL)
        return -1;

    p = en->m_recorded;
    frag_sum = 0L;
    u = 0;
    while(p) {
        frag_sum += p->len;
        Trace("  %4d: p %p next 0x%08x | len %4d/%d %7ld\n",
            u, p, p->next, p->len, p->size, (long)frag_sum);
        if(p==p->next) {
            Trace("  looped !!!\n");
            break;
        }
        p = p->next;
        u++;
    }
    Trace("  num %d, totalength %lld\n",
        u, (long long)en->totallength);

    return u;
}


off_t
skcache_show_with_HTML(int ofd)
{
    int i;
    off_t wsum, w;
    skiplist_t      *list;
    skiplist_cell_t *cur;
    skcache_entry_t *en;

    wsum = 0;

    w = fdprintf(ofd, "<html>\n");
    wsum += w;



    for(i=0;i<skcache_bucket_num;i++) {
        w = fdprintf(ofd, "<p>bucket# %d\n", i);
        wsum += w;

        w = fdprintf(ofd, "<listing>\n");
        wsum += w;

        list = &skcache_bucket_base[i].cup;
        cur = list->header->skip[0];

        while(cur && cur!= list->term) {
            en = (skcache_entry_t*)cur->entity;
            w = fdprintf(ofd, " %04d:%04d:%04d %d %d '%s'\n",
                    en->h1, en->h2, en->h3, en->ref, en->count, en->name);
            if(w>0) {
                wsum += w;
            }

            cur = cur->skip[0];
        }

        w = fdprintf(ofd, "</listing>\n");
        wsum += w;

    }


    w = fdprintf(ofd, "<html>\n");
    wsum += w;

    Trace("write %ld bytes\n", wsum);
}


int
skcache_have(int fh, int sh, int th, int ff, char *name,
    int flag_check_rcode, time_t ref_time)
{
    int ret;
    skcache_entry_t tmp;
    skcache_entry_t *stored_pos;
    bucket_t *bk;


    ret = 0;

    bk = &skcache_bucket_base[fh];

    tmp.h1 = fh;
    tmp.h2 = sh;
    tmp.h3 = th;
    tmp.h4 = ff;
    strcpy(tmp.name, name);

	LOCK_BK(bk);

    stored_pos = (skcache_entry_t*) sk_find(&bk->cup, &tmp, skcache_entry_cmp);

    if(stored_pos) {
#if 1
        if(strcmp((const char*)stored_pos->name, (const char*)name)!=0) {
            Error("skcache_have: hash collision %04d:%04d:%04d '%s' v.s. '%s'",
                fh, sh, th, stored_pos->name, name);
            goto out;
        }
#endif

        if(flag_check_rcode) {
        }

        if(ref_time>0) {
        }

        ret = 1;
    }
    else {
        ret = 0;
    }

out:
	UNLOCK_BK(bk);

    return ret;
}


static
int
alloc_cache_entry_freelist(skcache_entry_t** head, int num)
{
    char *base;
    int esize;
    int i;
    int ret;
    skcache_entry_t *pos;
    skcache_entry_t *last;

    esize = sizeof(skcache_entry_t);
    base = (char*)malloc(esize * num);
    if(base==NULL) {
        return 0;
    }

    for(i=0; i<num; i++) {
        pos = (skcache_entry_t*)(base + esize*i);

        pos->magic[0] = MAGIC_0;
        pos->magic[1] = MAGIC_1;

        pos->oid = -1;
		pos->ref = -1;
        pos->h1	= -1;
        pos->h2 = -1;
        pos->h3 = -1;
        pos->name[0] = '\0';
        pos->m_recorded = NULL;
        pos->d_recorded = -1;

        pos->next = (skcache_entry_t*)((char*)pos + esize);
    }

    *head = (skcache_entry_t*)base;
    pos->next = NULL;

#if 0

    Trace("entry// head %p, num %d\n", head, num);
    Trace("esize %d, base %p\n", esize, base);

    if(num<100) {
        i = 0;
        pos = *head;
        while(pos) {
            Trace(" %2d: pos %p, name '%s', m_recorded %p ->next %p\n",
                i, pos, pos->name, pos->m_recorded, pos->next);

            pos = pos->next;
            i++;
        }
    }

#endif

    return esize*num;
}

static
int
alloc_object_frag_freelist(skcache_frag_t** head, int num, int sz)
{
    char *base;
    int esize;
    int i;
    skcache_frag_t *pos;
    skcache_frag_t *last;

    esize = sizeof(skcache_frag_t)+sz;
    base = (char*)malloc(esize * num);
    if(base==NULL) {
        return 0;
    }

    for(i=0; i<num; i++) {
        pos = (skcache_frag_t*)(base + esize*i);
        pos->size = sz;
        pos->len = 0;
        pos->ent = (char*)((char*)pos+sizeof(skcache_frag_t));

        pos->next = (skcache_frag_t*)((char*)pos+esize);
    }
    *head = (skcache_frag_t*)base;
    pos->next = NULL;

#if 0
    Trace("frag// head %p, num %d, sz %d\n", head, num, sz);
    Trace("esize %d, base %p\n", esize, base);

    i = 0;
    pos = *head;
    while(pos) {
        if(i>num-3) {
            Trace(" %2d: pos %p, len %d, sz %d ->next %p\n",
                i, pos, pos->len, pos->size, pos->next);
        }

        pos = pos->next;
        i++;
    }

#endif

#if 0
    Trace("frag// head %p, num %d, sz %d\n", head, num, sz);
    Trace("esize %d, base %p\n", esize, base);

    if(num<100) {
        i = 0;
        pos = *head;
        while(pos) {
            Trace(" %2d: pos %p, len %d, sz %d ->next %p\n",
                i, pos, pos->len, pos->size, pos->next);

            pos = pos->next;
            i++;
        }
    }

#endif

    return esize*num;
}



int
skcache_entry_purge(bucket_t *bk, skcache_entry_t *e)
{
    skcache_frag_t *cur, *pre;

#if 0
    Trace("skcache_entry_purge: bk %p, e %p\n", bk, e);
#endif

    if(e->m_recorded) {
        /* remove */
        cur = e->m_recorded;
        while(cur) {
            pre = cur->next;

            skcache_push_idle_frag(bk, cur);
            if(pre==cur) {
                Error("skcache_entry_purge: bk %p, entry %p, frag %p is looped",
                    bk, e, cur);
                break;
            }

            cur = pre;
        }

        e->m_recorded = NULL;
    }

    /* XXX */
#if 0
    if(e->d_recorded) {
    }
#endif

#if 0
    skcache_push_idle_entry(bk, e);
#endif

    return 0;
}



int
skcache_entry_delete(bucket_t *bk, skcache_entry_t *e)
{
    void *chkv;
	int chk;

	Trace("skcache_entry_delete: bk# %d %p, entry %p\n",
		bk->bid, bk, e);

	chk = skcache_entry_purge(bk, e);
	if(chk) {
		Error("fail skcache_entry_purge()");
	}
	else {
		chkv = sk_delete(&bk->cup, e,
			&bk->elem_freelist, skcache_entry_cmp, NULL, 0);
		if(chkv==NULL) {
			Error("skcache_entry_delete: fail sk_delete() chkv %p", chkv);
			return 1;
		}

		chk = skcache_push_idle_entry(bk, e);
		Trace("skcache_entry_delete: chk %d\n", chk);
	}

    return 0;
}

#if 0

int
skcache_entry_delete(bucket_t *bk, skcache_entry_t *e)
{
	int chk;

    chk = sk_delete(&bk->cup, e,
        &bk->elem_freelist, skcache_entry_cmp, NULL, 0);

    skcache_push_idle_entry(bk, e);

    if(chk==NULL) {
        return 1;
    }
    return 0;
}

#endif


char* cache_dir_prefix = "/var/tmp/cache";

inline
char*
assign_filename(skcache_entry_t *e, char *name)
{
#if 0
    sprintf(name, "%s/%d/%d/%d/%d",
        cache_dir_prefix, e->h1, e->h2, e->h3, e->oid);
#endif
#if 0
    sprintf(name, "%s/%d-%d-%d-%d",
        cache_dir_prefix, e->h1, e->h2, e->h3, e->oid);
#endif
#if 0
    sprintf(name, "%s/%d/%d-%d-%d",
        cache_dir_prefix, e->h1, e->h2, e->h3, e->oid);
#endif
#if 1
    sprintf(name, "%s/%d/%d/%d/%d",
        cache_dir_prefix, e->h1, e->h2, e->h3, e->h4);
#endif

#if 0
    Trace("  name '%s'\n", name);
#endif

    return name;
}

inline
char*
assign_dirname(skcache_entry_t *e, char *name, int level)
{
    switch(level) {
    case 0:
        sprintf(name, "%s",
            cache_dir_prefix);
        break;
    case 1:
        sprintf(name, "%s/%d",
            cache_dir_prefix, e->h1);
        break;
    case 2:
        sprintf(name, "%s/%d/%d",
            cache_dir_prefix, e->h1, e->h2);
        break;
    default:
        sprintf(name, "%s/%d/%d/%d",
            cache_dir_prefix, e->h1, e->h2, e->h3);
        break;
    }

    return name;
}

inline
char*
assign_dirname_bynumber(int d1, int d2, int d3, char *name)
{
	if(d2<0) {
    	sprintf(name, "%s/%d", cache_dir_prefix, d1);
	}
	else
	if(d3<0) {
    	sprintf(name, "%s/%d/%d", cache_dir_prefix, d1, d2);
	}
	else {
    	sprintf(name, "%s/%d/%d/%d", cache_dir_prefix, d1, d2, d3);
	}
    return name;
}


#ifndef HAVE_DIRENT_T
typedef struct dirent dirent_t;
#endif

int
build_dir(skcache_entry_t *e, int lev)
{
    int i;
    int ret=0;
    int chk;
    char tmp[L_PATH];

    for(i=1;i<=lev;i++) {
        assign_dirname(e, tmp, i);
        mkdir(tmp, 0755);
        if(chk && errno!=EEXIST) {
            Error("build_dir fail mkdir() '%s' (%d)", tmp, errno);
            ret = -1;
            goto out;
        }
    }

out:
    return ret;
}

int
build_dir_bynumber(int d1, int d2, int d3)
{
    int i;
    int ret=0;
    int chk;
    char name[L_PATH];

l1:
    sprintf(name, "%s/%d", cache_dir_prefix, d1);
    mkdir(name, 0755);
    if(chk && errno!=EEXIST) {
        Error("build_dir_bynumber fail mkdir() '%s' 1 (%d)", name, errno);
        ret = -1;
        goto out;
    }

l2:
    sprintf(name, "%s/%d/%d", cache_dir_prefix, d1, d2);
    mkdir(name, 0755);
    if(chk && errno!=EEXIST) {
        Error("build_dir_bynumber fail mkdir() '%s' 2 (%d)\n", name, errno);
        ret = -1;
        goto out;
    }

l3:
    sprintf(name, "%s/%d/%d/%d", cache_dir_prefix, d1, d2, d3);
    mkdir(name, 0755);
    if(chk && errno!=EEXIST) {
        Error("build_dir_bynumber fail mkdir() '%s' 3 (%d)\n", name, errno);
        ret = -1;
        goto out;
    }

out:
    return ret;
}


/*
 * return
 *      0   success
 *      >0  fail
 *
 */

int
skcache_entry_swapload(bucket_t *bk, skcache_entry_t *e)
{
    int sfd;
    char name[L_PATH];
    skcache_entry_t tmp_entry;
    skcache_frag_t *cur, *pre;
    int nr;
    off_t nr_sum;
static int x=0;
    int chk;
    char *namechk;
    int i;
    int ret;
    struct timeval begin, end;
	int c;
	off_t fsize;
	int erase;


    gettimeofday(&begin, NULL);

    MUTEX_LOCK(&stat_lock);
    swapload_count++;
    MUTEX_UNLOCK(&stat_lock);

    ret = 0;

#if 0
    Trace("skc_swapload: p %p, oid# %d, bk# %d, %p\n", e, e->oid, bk->bid, bk);
#endif
#if 0
    Trace("skc_swapload: p %p, id# %d, bk %p, frag = %4d/%d, len %ld (%% %d)\n",
        e, e->oid,
        bk, bk->frag_use, bk->frag_all,
        (long)e->totallength, e->totallength%1460);
#endif
#if 0
    if(e->d_recorded<0) {
        Error("not recorded in disk");
        ret = 1;
        goto out;
    }
#endif
    if(e->m_recorded!=NULL) {
        Error("swapload: recorded in memory");
        ret = 2;
        goto out;
    }
    if(bk->frag_use>bk->frag_all) {
        Error("swapload: no frag in freelist");
        ret = 3;
        goto out;
    }

    if(bk->dirty2 == e->h2 && bk->dirty3 == e->h3) {
        Error("swapload: conflict directory [%03d:%03d:%03d] to sweeper",
            e->h1, e->h2, e->h3);
        ret = 73;
        goto out;
    }

    namechk = assign_filename(e, name);

#if 0
    Trace("  name '%s'\n", name);
#endif
    sfd = open(name, O_RDONLY, 0666);
    if(sfd<0) {
#if 0
        Trace("swapload: cannot open swapload file '%s' (%d)\n", name, errno);
#endif
#if 0
        Error("cannot open swapload file '%s' (%d)", name, errno);
#endif
        ret = 4;
        goto out;
    }

	fsize = (off_t) lseek(sfd, 0, SEEK_END);
	if(fsize>(off_t)0) {
		lseek(sfd, 0, SEEK_SET);
	}

	c = 0;
read_cntl:
	c++;
    nr = read(sfd, &tmp_entry, sizeof(skcache_entry_t));
    if(nr<0) {
        if(errno==EINTR && c < DISK_RETRY) {
            goto read_cntl;
        }
        Error("swapload: cntl, fail read() return %d %c (%d)", nr, c, errno);
        ret = 123;
        goto fail_read_cntl;
    }

	if(nr<sizeof(skcache_entry_t)) {
#if 0
        Trace("swapload: to short %ld bytes\n", (long)nr);
#endif
		ret = 126;
        goto fail_read_cntl;
	}

    if(tmp_entry.magic[0]!=MAGIC_0
        || tmp_entry.magic[1]!=MAGIC_1) {
		Error("swapload: ignore magic %02x %02x / %02x %02x '%s'",
			tmp_entry.magic[0], tmp_entry.magic[1],
			MAGIC_0, MAGIC_1, name);
        ret = 125;
        goto fail_read_cntl;
    }

#if 0
    Trace("swapload: read oid# %d, hash %d,%d,%d,%d %s\n",
        tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
        tmp_entry.h3, tmp_entry.h4, tmp_entry.name);
#endif

    if(tmp_entry.oid<0) {
#if 0
        Trace("swapload: reject negative oid to %p\n", e);
#endif
        ret = 127;
        goto fail_read_cntl;
    }

	if(strcmp(tmp_entry.name,e->name)!=0) {
#if 1
		Trace("swapload: swapload file '%s'\n", name);
		Trace("swapload: read oid# %d, hash %d,%d,%d,%d %s\n",
			tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
			tmp_entry.h3, tmp_entry.h4, tmp_entry.name);
		Trace("but target is '%s'\n", e->name);
#endif
		ret = 133;
		erase = 1;
		goto fail_read_cntl;
	}

	if(tmp_entry.issue_time<0) {
#if 1
		Error("swapload: ignore issue_time\n");
#endif
        ret = 134;
        goto fail_read_cntl;
	}

    if(tmp_entry.issue_time<min_date) {
#if 1
		Error("swapload: very old object\n");
#endif
        ret = 129;
        goto fail_read_cntl;
    }

    /* XXX */
    *e = tmp_entry;
    e->m_recorded = NULL;

skip_read_cntl:
    (void)1;

    pre = NULL;

    nr_sum = 0L;
    i = 0;
    while(cur) {
        if(bk->frag_freelist==NULL) {
            Error("swapload: no free frag");
            ret = 5;
            goto fail_read;
        }

        cur = skcache_pull_new_frag(bk);
        if(cur==NULL) {
            ret = 7;
            goto fail_read;
        }

		c=0;
try_read:
		c++;
        nr = read(sfd, cur->ent, cur->size);
#if 0
        Trace("<<< %3d %p: read %5d/%ld %7d?\n",
            i, cur, nr, (long)cur->size,
            (nr>0) ? (long)nr_sum+nr : (long)nr_sum);
#endif
        if(nr<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_read;
            }

            cur->len = 0;
            Error("swapload: data, fail read() return %d %d (%d)", nr, c, errno);
            ret = 6;

            goto fail_read;
        }
        if(nr==0) {
            cur->len = 0;
            break;
        }

        cur->len = nr;
        nr_sum += nr;

        if(pre==NULL) {
            e->m_recorded = cur;
        }
		else {
			pre->next = cur;
		}

        i++;
        pre = cur;
		cur->next = NULL; 	/* dummy */
    }

    close(sfd);
    sfd = -1;

#if 0
    Trace("  swapload %ld bytes\n", (long)nr_sum);
#endif

    e->d_recorded = -1;


#if 0
    Trace("  swapload read %ld/%ld '%s'\n",
        (long)nr_sum, (long)e->totallength, name);
#endif


    cur = e->m_recorded;
#if 0
    if(cur)
    {
        int hlen;
        int xvmajor, xvminor;
        time_t xdate, xlmd, xexpire;
        off_t xclen;
        int xrcode;
        int xcachable;
        int xreload;

        hlen = HTTP_find_header_term(cur->ent, MIN(cur->len, e->totallength));

        chk = HTTP_parse_response_header(cur->ent, MIN(hlen, cur->len),
                &xvmajor, &xvminor, &xrcode,
                &xdate, &xlmd, &xexpire, &xclen,
                &xcachable, &xreload);

        if(chk) {
            Error("  *ignore* header");
        }

        Trace("  hlen %ld, xclen %ld + 1 -> %ld vs len %ld, diff %ld\n",
            (long)hlen, (long)xclen,
            (long)(xclen + hlen +1),
            (long)e->totallength,
            (long)(xclen + hlen +1 - e->totallength));

    }
#endif

    MUTEX_LOCK(&stat_lock);
    swapload_success++;
    MUTEX_UNLOCK(&stat_lock);

#if 1
    Trace("swapload: success p %p, oid# %d, bk# %d, %p\n", e, e->oid, bk->bid, bk);
#endif

    goto out;

fail_read_cntl:
    close(sfd);

#if 1
    Trace("swapload: fail read contorl block reason %d, erase %d, size %ld\n",
	ret, erase, (long)fsize);
#endif
    if(erase) {
	unlink(name);
	MUTEX_LOCK(&stat_lock);
	swap_count--;
	swap_size -= fsize;
	MUTEX_UNLOCK(&stat_lock);
    }

    goto out;

fail_read:
    close(sfd);

#if 1
    Trace("swapload: fail to read data. trush just readings\n");
#endif
	if(e->m_recorded) {
		cur = e->m_recorded;
		while(cur) {
			pre = cur->next;

			skcache_push_idle_frag(bk, cur);
			if(pre==cur) {
				break;
			}

			cur = pre;
		}
		e->m_recorded = NULL;
	}

    goto out;

out:
#if 0
    x++;
    Trace("` %d\n", x);
    if(x>1280) exit(17);
#endif

#if 0
    Trace("  frag use/all = %d/%d\n", bk->frag_use, bk->frag_all);
#endif

    gettimeofday(&end, NULL);
    sim_stat_add_value(&swapload_time, KOTETU_SUB_TIME(end, begin));

#if 0
    Trace("swapload: return %d\n", ret);
#endif

    return ret;
}




int
skcache_entry_swapin(bucket_t *bk, skcache_entry_t *e)
{
    int sfd;
    char name[L_PATH];
    skcache_entry_t tmp_entry;
    skcache_frag_t *cur, *pre;
    int nr;
    off_t nr_sum;
static int x=0;
    int chk;
    char *namechk;
    int i;
    int ret;
    struct timeval begin, end;
	int c;

 assert(check_bkentry(bk, "si0"));

    gettimeofday(&begin, NULL);

    MUTEX_LOCK(&stat_lock);
    swapin_count++;
    MUTEX_UNLOCK(&stat_lock);

    ret = 0;

#if 0
    Trace("skc_swapin: p %p, id# %d, bk %p, frag = %4d/%d, len %ld (%% %d)\n",
        e, e->oid,
        bk, bk->frag_use, bk->frag_all,
        (long)e->totallength, e->totallength%1460);
#endif
    if(e->d_recorded<0) {
#if 0
        Error("not recorded in disk");
#endif
#if 0
        ret = 1;
        goto out;
#endif
    }
    if(e->m_recorded!=NULL) {
        Error("swapin: recorded in memory");
        ret = 2;
        goto out;
    }
    if(bk->frag_use>bk->frag_all) {
        Error("swapin: no frag in freelist");
        ret = 3;
        goto out;
    }

    if(bk->dirty2 == e->h2 && bk->dirty3 == e->h3) {
        Error("swapin: conflict directory [%03d/%03d/%03d] to sweeper",
            e->h1, e->h2, e->h3);
        ret = 73;
        goto out;
    }

    namechk = assign_filename(e, name);

#if 0
    Trace("  name '%s'\n", name);
#endif
    sfd = open(name, O_RDONLY, 0666);
    if(sfd<0) {
#if 0
        Error("cannot open swapin file '%s' (%d)", name, errno);
#endif
        ret = 4;
        goto out;
    }

 assert(check_bkentry(bk, "si2"));

	c = 0;
read_cntl:
	c++;
    nr = read(sfd, &tmp_entry, sizeof(skcache_entry_t));
    if(nr<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto read_cntl;
        }
        Error("fail read() return %d %d (%d)", nr, c, errno);
        ret = 123;
        goto fail_read_cntl;
    }

#if 1
	Trace("e %p, tmp_entry %p\n",
		e, tmp_entry);
#endif

    if(tmp_entry.magic[0]!=MAGIC_0
        || tmp_entry.magic[1]!=MAGIC_1) {
		Error("swapin: ignore magic %02x %02x / %02x %02x '%s'",
			tmp_entry.magic[0], tmp_entry.magic[1],
			MAGIC_0, MAGIC_1, name);
        ret = 125;
        goto fail_read_cntl;
    }

 assert(check_bkentry(bk, "si2.5"));
#if 1
    Trace("swapin: given oid# %d, hash %d,%d,%d,%d %s\n",
        e->oid, e->h2, e->h2,
        e->h3, e->h4, e->name);
    Trace("swapin: read  oid# %d, hash %d,%d,%d,%d %s\n",
        tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
        tmp_entry.h3, tmp_entry.h4, tmp_entry.name);
#endif

    if(tmp_entry.oid<0) {
#if 0
        Trace("swapin: reject negative oid to %p\n", e);
#endif
        ret = 127;
        goto fail_read_cntl;
    }
    if(tmp_entry.issue_time<min_date) {
        ret = 129;
        goto fail_read_cntl;
    }
#if 1
	
    Trace("  elem  %4d/%-5d //\n",
        bk->cup.num, bk->elemnum);

    Trace("  entry %4d/%-5d // hh %4d h %4d l %4d : %p\n",
        bk->entry_use, bk->entry_all,
        bk->entry_hhwater, bk->entry_hwater, bk->entry_lwater,
        bk->entry_freelist);

    Trace("  frag  %4d/%-5d // hh %4d h %4d l %4d : %p\n",
        bk->frag_use, bk->frag_all,
        bk->frag_hhwater,  bk->frag_hwater, bk->frag_lwater,
        bk->frag_freelist);

#endif

 assert(check_bkentry(bk, "si2.8"));


	dump_bkentryfreelist(bk, "before");

    /* XXX */
#if 0
    *e = tmp_entry;
#else
	memcpy(e, &tmp_entry, sizeof(skcache_entry_t));
#endif
    e->m_recorded = NULL;
    e->next = NULL;

	dump_bkentryfreelist(bk, "after ");

 assert(check_bkentry(bk, "si3"));

skip_read_cntl:
    (void)1;

    pre = NULL;

    nr_sum = 0L;
    i = 0;
    while(cur) {
        if(bk->frag_freelist==NULL) {
            Error("swapin: no free frag");
            ret = 5;
            goto fail_read;
        }

        cur = skcache_pull_new_frag(bk);
        if(cur==NULL) {
            ret = 7;
            goto fail_read;
        }

		c = 0;
try_read:
		c++;
        nr = read(sfd, cur->ent, cur->size);
#if 0
        Trace("<<< %3d %p: read %5d/%ld %7d?\n",
            i, cur, nr, (long)cur->size,
            (nr>0) ? (long)nr_sum+nr : (long)nr_sum);
#endif
        if(nr<0) {
            if(errno==EINTR && c< DISK_RETRY) {
                goto try_read;
            }

            cur->len = 0;
            Error("fail read() return %d %d (%d)", nr, c, errno);
            ret = 6;

            goto fail_read;
        }
        if(nr==0) {
            cur->len = 0;
            break;
        }

        cur->len = nr;
        nr_sum += nr;

        if(pre==NULL) {
            e->m_recorded = cur;
        }
        else {
            pre->next = cur;
        }

        i++;
        pre = cur;
		cur->next = NULL;
    }

    close(sfd);
    sfd = -1;

#if 0
    Trace("  swapin %ld bytes\n", (long)nr_sum);
#endif

    e->d_recorded = -1;


#if 0
    Trace("  swapin read %ld/%ld '%s'\n",
        (long)nr_sum, (long)e->totallength, name);
#endif

 assert(check_bkentry(bk, "si4"));

    cur = e->m_recorded;
#if 0
    if(cur)
    {
        int hlen;
        int xvmajor, xvminor;
        time_t xdate, xlmd, xexpire;
        off_t xclen;
        int xrcode;
        int xcachable;
        int xreload;

        hlen = HTTP_find_header_term(cur->ent, MIN(cur->len, e->totallength));

        chk = HTTP_parse_response_header(cur->ent, MIN(hlen, cur->len),
                &xvmajor, &xvminor, &xrcode,
                &xdate, &xlmd, &xexpire, &xclen,
                &xcachable, &xreload);

        if(chk) {
            Error("  *ignore* header");
        }

        Trace("  hlen %ld, xclen %ld + 1 -> %ld vs len %ld, diff %ld\n",
            (long)hlen, (long)xclen,
            (long)(xclen + hlen +1),
            (long)e->totallength,
            (long)(xclen + hlen +1 - e->totallength));

    }
#endif

    MUTEX_LOCK(&stat_lock);
    swapin_success++;
    MUTEX_UNLOCK(&stat_lock);

    goto out;

fail_read_cntl:
    close(sfd);

#if 0
    Trace("swapin: fail read contorl block\n");
#endif

    goto out;

fail_read:
    close(sfd);

 assert(check_bkentry(bk, "si5"));

#if 0
    Trace("swapin: trush just reads\n");
#endif
    cur = e->m_recorded;
    while(cur) {
        pre = cur->next;

        skcache_push_idle_frag(bk, cur);
        if(pre==cur) {
            break;
        }

        cur = pre;
    }

 assert(check_bkentry(bk, "si6"));
    goto out;

out:
#if 0
    x++;
    Trace("` %d\n", x);
    if(x>1280) exit(17);
#endif

#if 0
    Trace("  frag use/all = %d/%d\n", bk->frag_use, bk->frag_all);
#endif

    gettimeofday(&end, NULL);
    sim_stat_add_value(&swapin_time, KOTETU_SUB_TIME(end, begin));


 assert(check_bkentry(bk, "si9"));

    return ret;
}








#define XSK_DISKFULL (-27)

/*
 *
 * return
 *		0	success
 *		!0	fail
 */

int
skcache_entry_swapout(bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int sfd;
    char name[L_PATH];
    skcache_frag_t *cur, *pre;
    int nw;
    off_t nw_sum;
static int x=0;
    int chk;
    char *namechk;
    int i;
    struct timeval begin, end;
    struct stat sbuf;
    int sig_pos;
	int c;

    off_t last_size;


    gettimeofday(&begin, NULL);

    ret = 0;


	if(bk==NULL || e==NULL) {
		Error("nil bucket pointer %x or entry pointer %x", bk, e);
		ret = -789;
		goto out;
	}

#if 0
    Trace("skc_swapout: p %p, id# %d, bk %p, frag = %4d/%d, len %ld (%% %d)\n",
        e, e->oid,
        bk, bk->frag_use, bk->frag_all,
        (long)e->totallength, e->totallength%1460);
#endif
    if(e->d_recorded>=0) {
        Error("already recorded to disk, escape");
        ret = 1;
        goto out;
    }
    if(e->m_recorded==NULL) {
        Error("swapout: not recorded in memory");
        ret = 2;
        goto out;
    }
#if 1
	if(e->magic[0]!=MAGIC_0 || e->magic[1]!=MAGIC_1) {
		Error("swapout: strange magic %02x %02x / %02x %02x",
			e->magic[0], e->magic[1], MAGIC_0, MAGIC_1);	

		ret = 4;
		goto out;
	}
#endif

    if(bk->dirty2 == e->h2 && bk->dirty3 == e->h3) {
        Error("swapout: conflict directory [%03d:%03d:%03d] to sweeper",
            e->h1, e->h2, e->h3);
        ret = 3;
        goto out;
    }


    MUTEX_LOCK(&stat_lock);
	chk = swapout_token;
	if(chk>0) {
		swapout_token--;
    	swapout_count++;
	}
    MUTEX_UNLOCK(&stat_lock);

	if(chk<=0) {
#if 0
		Trace("swapout: no more token. wait several tick\n");
#endif
		ret = -999;
        goto out;
	}
	if(chk==1) {
		Trace("swapout: *last* token of bk# %d %p\n", bk->bid, bk);
	}


    cur = e->m_recorded;
#if 0
    {
        int hlen;
        int xvmajor, xvminor;
        time_t xdate, xlmd, xexpire;
        off_t xclen;
        int xrcode;
        int xcachable;
        int xreload;

        hlen = HTTP_find_header_term(cur->ent, MIN(cur->len, e->totallength));

        chk = HTTP_parse_response_header(cur->ent, MIN(hlen, cur->len),
                &xvmajor, &xvminor, &xrcode,
                &xdate, &xlmd, &xexpire, &xclen,
                &xcachable, &xreload);

        if(chk) {
            Error("  *ignore* header");
        }

#if 0
        if((xclen + hlen +1 - e->totallength)!=0) { }
#endif
        Trace("  hlen %ld, xclen %ld + 1 -> %ld vs len %ld, diff %ld\n",
            (long)hlen, (long)xclen,
            (long)(xclen + hlen +1),
            (long)e->totallength,
            (long)(xclen + hlen +1 - e->totallength));

    }
#endif


    namechk = assign_filename(e, name);
#if 0
    Trace("  name '%s'\n", name);
#endif

	last_size = -1;

	c = 0;
try_open:
	c++;
/*
    sfd = open(name, O_WRONLY | O_CREAT, 0666);
*/
    sfd = open(name, O_RDWR | O_CREAT, 0666);
    if(sfd<0) {
		if(errno==EINTR && c<DISK_RETRY) {
			goto try_open;
		}
#if 1
        if(errno==ENOENT && c<DISK_RETRY) {
#if 1
			Trace("swapout: setup new dir %d,%d,%d and open '%s'\n",
				e->h1, e->h2, e->h3, name);
#endif
           	build_dir_bynumber(e->h1, e->h2, e->h3);
            goto try_open;
        }
#endif
		if(errno==ENOSPC) {
        	Error("swapout: disk full on open()");
#if 0
			MUTEX_LOCK(&stat_lock);
			swapout_token = 0;
			MUTEX_UNLOCK(&stat_lock);
#endif

			ret = XSK_DISKFULL;
			goto out;
		}
       	Error("cannot open swapout file '%s' %d (%d)", name, c, errno);

        ret = 4;
        goto out;
    }

	last_size = (off_t)lseek(sfd, 0, SEEK_END);
    if(sfd<0) {
        Error("fail lseek() swap out file '%s' (%d)", name, errno);
        ret = 17;
		close(sfd);
        goto out;
	}
	if(last_size!=(off_t)0) {
		lseek(sfd, 0, SEEK_SET);
	}
#if 0
Trace("swapout: oid# %d, last_size %ld\n", e->oid, (long)last_size);
#endif


	c = 0;
try_write_cntl:
	c++;
    nw = write(sfd, e, sizeof(skcache_entry_t));
    if(nw<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_write_cntl;
        }

        if(errno==ENOSPC) {
            Error("disk full!! at write control in swapout");
        }
		else
		if(errno==ENOMEM) {
            Error("kernel said no memory. reset token <%d>", c);

    MUTEX_LOCK(&stat_lock);
	swapout_token = 0;
    MUTEX_UNLOCK(&stat_lock);

{
            struct timespec w;

            w.tv_sec = 0;
            w.tv_nsec = 50 * 1000000;  /* 50 msec */

            nanosleep(&w, NULL);
}

			if(c<DISK_RETRY) {
				goto try_write_cntl;
			}

		}
		else {
        	Error("fail write() return %d %d (%d)", nw, c, errno);
		}
        ret = 5;
        goto fail_write_cntl;
    }

#if 0
{
	char buf[L_BUF];
	int nr;

	lseek(sfd, 0, SEEK_SET);
	nr = read(sfd, buf, nw);
	if(nr>0) {
		if(nr!=nw) {
			Error("read %d. written %d %s", nr, nw, name);
		}
		else
	    if(buf[0]!=MAGIC_0 || buf[1]!=MAGIC_1) {
			Error("swapout: strange written magic %02x %02x / %02x %02x",
				buf[0], buf[1], MAGIC_0, MAGIC_1);
		}
		else {
			Trace("swapout: ok '%s'\n", name);
		}
	}
	else {
			Trace("swapout: error or empty '%s' (%d)\n", name, errno);
	}
	
	lseek(sfd, nw, SEEK_SET);
}
#endif


    nw_sum = 0L;
    i = 0;
    while(cur) {
        if(cur->len>0) {

			c = 0;
try_write:
			c++;
            nw = write(sfd, cur->ent, cur->len);

#if 0
            Trace(">>> %3d %p: stored %5d/%d, write %5d %7ld?\n",
                i, cur->ent, cur->len, cur->size, nw,
                nw>0 ? (long)nw_sum+nw : (long)nw_sum);
#endif

            if(nw<0) {
                if(errno==EINTR && c<DISK_RETRY) {
                    goto try_write;
                }
                Error("fail write() return %d (%d)", nw, errno);
                if(errno==ENOSPC) {
                    Error("disk full!! when swapout data");
                }

                ret = 7;
                goto fail_write;
            }
            else
            if(nw==0) {
                break;
            }
            if(nw>0) {
                nw_sum += nw;
            }
        }
#if 0
        else {
            Trace(">>> %3d %p: stored %5d/%d %7ld\n",
                i, cur->ent, cur->len, cur->size, (long)nw_sum);
        }
#endif

        i++;
        cur = cur->next;
    }
#if 0
    Trace("  name '%s'\n", name);
    Trace("  swap out %ld bytes '%s'\n", (long)nw_sum, name);
#endif

	ftruncate(sfd, (off_t) nw_sum + sizeof(skcache_entry_t) );
    close(sfd);
    sfd = -1;

    cur = e->m_recorded;
    while(cur) {
        pre = cur->next;

        skcache_push_idle_frag(bk, cur);
        if(pre==cur) {
            Error("swapout: bk %p, entry %p, frag %p is looped",
                bk, e, cur);
            break;
        }

        cur = pre;
    }

    e->m_recorded = NULL;
    e->d_recorded = e->oid;

#if 0
    Trace("swapout: oid# %d, write %ld, %ld+%ld/%ld '%s'\n",
		e->oid, sizeof(skcache_entry_t) + (long)nw_sum,
		sizeof(skcache_entry_t), (long)nw_sum, (long)e->totallength, name);
#endif

    sig_pos = e->h1*radix_h2*radix_h3
        + e->h2*radix_h3 + e->h3;


    MUTEX_LOCK(&stat_lock);
    swapout_success++;

    if(last_size<=0) {
        swap_count += 1LL;
        swap_size +=
			(long long)(nw_sum + (off_t)sizeof(skcache_entry_t));
    }
    else {
        swap_size +=
			(long long)(nw_sum + (off_t)sizeof(skcache_entry_t) - last_size);
    }
#if 0
    Trace("swapout: oid# %d, total swap count %lld size %lld\n",
		e->oid, swap_count, swap_size);
#endif

    if(dir_sig[sig_pos]<0) {
        /* nothing */
    }
    else {
        dir_sig[sig_pos] |= e->h4;
    }


    MUTEX_UNLOCK(&stat_lock);

    goto out;

fail_write_cntl:
fail_write:
    close(sfd);

    unlink(name);

    MUTEX_LOCK(&stat_lock);
    if(last_size<0) {
    }
    else {
        swap_size -= last_size;
    }
    MUTEX_UNLOCK(&stat_lock);

    goto out;

out:
#if 0
    x++;
    Trace("` %d\n", x);
    if(x>1280) exit(17);
#endif
#if 0
    Trace("  frag use/all = %d/%d\n", bk->frag_use, bk->frag_all);
#endif

    gettimeofday(&end, NULL);
    sim_stat_add_value(&swapout_time, KOTETU_SUB_TIME(end, begin));

    return ret;
}





int
skcache_entry_purge_check_noref(void *pp, void *vv)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    bucket_t *bk = (bucket_t*)vv;
    int ret = 0;

#if 0
    Trace("  entry_check_noref: p %p ref %d in bk%p ", p, p->ref, bk);
#endif

    if(p->ref>0) {
#if 0
        Trace("refered %d session, not purge\n", p->ref);
#endif
        ret = 0;
        goto skip;
    }
    else {
#if 0
        Trace("no refered , *purge*\n");
#endif
        skcache_entry_purge(bk, p);
        skcache_push_idle_entry(bk,  p);
        p = NULL;

        ret = 1;
    }

skip:
    return ret;

}



int
skcache_entry_purge_check(void *pp, void *vv)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    bucket_t *bk = (bucket_t*)vv;
    int ret = 0;
    time_t now;
    int elap, idle;
    int erase;

    erase = 0;

    now = time(NULL);
    if(p->issue_time > 0 ) {
        elap = now - p->issue_time;
    }
    else {
#if 0
        Trace("pc: empty issue_time\n");
#endif
        elap = skcache_mem_keep_max;
    }
    if(p->hot > 0 ) {
        idle = now - p->hot;
    }
    else {
#if 0
        Trace("empty hot\n");
#endif
        idle = skcache_mem_keep_cur/2;
    }

    if(p->ref>0) {
        erase = 0;
    }
    else
    if(p->expire>0 && p->expire >= now) {
        erase = 1;
    }
    else
    if(idle >= skcache_mem_keep_cur) {
        erase = 1;
    }

#if 0
    Trace("  check %p, ref %d, freq %d, idle %4d/%d, elap %5d: ",
        p, p->ref, p->count, idle, skcache_mem_keep_cur, elap);
#endif

    if(erase>0) {
#if 0
        Trace("*purge*\n");
#endif
        skcache_entry_purge(bk, p);
        skcache_push_idle_entry(bk,  p);
        p = NULL;

        ret = 1;
    }
    else {
#if 0
        Trace("not purge\n");
#endif
        ret = 0;
    }

skip:
    return ret;

}


#ifdef NO_VS
#ifdef VS
#undef VS
#endif
#endif



inline
static
int
_skcache_entry_swapout_check(void *pp, void *vv, int force)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    bucket_t *bk = (bucket_t*)vv;
    int ret = 0;
    time_t now;
    int elap, idle;
    int action;
	int chk;

    action = 0;

    now = time(NULL);

    if(p->issue_time > 0 ) {
        elap = now - p->issue_time;
    }
    else {
#if 1
        Error("_sw: empty issue_time entry %p", p);
#endif
        elap = -1;
    }

    if(p->hot > 0 ) {
        idle = now - p->hot;
    }
    else {
#if 0
        Trace("empty hot\n");
#endif
#if 0
        idle = MIN(skcache_mem_keep_cur/2, elap);
#endif
        idle = -1;
    }

#ifdef VS
    if(p->m_recorded) {
    Trace(" %p:: %8d %7d %c 0x%8x %3d, c %d, f %2d, idle %4d, elap %5d: ",
        p, p->oid, p->totallength,
        (p->ref > 0) ? 'R' : '-',
        p->m_recorded, p->d_recorded, p->cracked, p->count,
        idle, elap);
    }
    else {
    Trace(" %p:: %8d %7d %c %5s %8d, c %d, f %2d, idle %4d, elap %5d: ",
        p, p->oid, p->totallength,
        (p->ref > 0) ? 'R' : ' ',
        "-", p->d_recorded, p->cracked, p->count,
        idle, elap);
    }
#endif

#if 0
	Trace("lmd %d now %d diff %d ", p->lmd, now, now - p->lmd);
#endif

    /* action
        3   swapout + purge
        2   purge
        1   swapout
        0   no action
     */

    if(p->ref>0) {
#ifdef VS
Trace("refered ");
#endif
        action = 0;
    }
    else
	if(elap<0) {
#ifdef VS
Trace("junk    ");
#endif
        action = 2;
	}
	else
    if(p->cracked) {
#ifdef VS
Trace("cracked ");
#endif
        action = 2;
    }
    else
    if(p->d_recorded>=0) {
        if(p->m_recorded==NULL) {
#ifdef VS
Trace("already ");
#endif
            /* already swapouted */
            action = 0;
        }
        else {
#ifdef VS
Trace("mine-01 ");
#endif
            action = 2;
        }
    }
    else
    if(p->m_recorded==NULL) {
#ifdef VS
Trace("empty-b ");
#endif
        action = 2;
    }
    else
    if(p->totallength<=(off_t)0) {
#ifdef VS
Trace("empty-c ");
#endif
        action = 2;
    }
	else
    if(elap > skcache_entry_keep_cur) {
#ifdef VS
Trace("expire1 ");
#endif
        action = 2;
	}
	else
    if(p->expire>0 && p->expire <= now) {
#ifdef VS
Trace("expire2 ");
#endif
        action = 2;
    }
#if 0
    else
    if(p->lmd>0 && now - p->lmd > skcache_entry_keep_cur) {
#ifdef VS
Trace("expire3 ");
#endif
        action = 2;
	}
#endif
    else {

  if(force) {
	
		if(idle < 0) {
#ifdef VS
Trace("neg,    ");
#endif
            action = 2;
		}
		else {
#ifdef VS
Trace("shift   ");
#endif
            action = 3;
        }

  }
  else {

#if 1
		/*	
		 * XXX
		 * randomize forward swapout feature
	 	 */
		if(idle<=skcache_mem_keep_min) {
			/* nothing */
		}
		else {
			idle += rand() % skcache_mem_keep_cur;
		}
#endif

	
		if(idle < 0) {
#ifdef VS
Trace("neg,    ");
#endif
            action = 2;
		}
		else
        if(idle >= skcache_mem_keep_cur) {
#ifdef VS
Trace("shift   ");
#endif
            action = 3;
        }
        else {
#ifdef VS
Trace("keep    ");
#endif
        }

  }


    }

	/* collect stat */
	if(action>=0 && action<4) {
		bk->op[action%4]++;
#ifdef VS
		Trace("op %d ", action);
#endif
	}

	ret = 0;
	switch(action) {
	case 3:
        chk = skcache_entry_swapout(bk, p);
		if(chk) {
            bk->inmemory++;
			ret = 0;
#ifdef VS
			Trace("*swapout-purge* fail (%d)\n", chk);
#endif

			if(chk==XSK_DISKFULL) {
				skcache_push_idle_entry(bk, p);
				p = NULL;
				ret = 1;
				break;
			}
		}
		else {
            bk->indisk++;
			skcache_push_idle_entry(bk, p);
			p = NULL;
			ret = 1;
#ifdef VS
			Trace("*swapout-purge* success\n");
#endif
		}
		break;

	case 2:
        chk = skcache_entry_purge(bk, p);
		if(chk) {
            bk->inmemory++;
			ret = 0;
#ifdef VS
			Trace("*purge* fail (%d)\n", chk);
#endif
		}
		else {
			skcache_push_idle_entry(bk, p);
			p = NULL;
			ret = 1;
#ifdef VS
			Trace("*purge*\n");
#endif
		}
		break;

	case 1:
        chk = skcache_entry_swapout(bk, p);
		if(chk) {
            bk->inmemory++;
#ifdef VS
			Trace("*swapout* fail (%d)\n", chk);
#endif
		}
		else {
            bk->inmemory++;
            bk->indisk++;
#ifdef VS
			Trace("*swapout* success\n");
#endif
		}

		break;

	default:
            bk->inmemory++;
#ifdef VS
        Trace("no-change\n");
#endif
		break;
    }

#if 0
    if(p) {
        if(p->m_recorded) {
            bk->inmemory++;
        }
        if(p->d_recorded>=0) {
            bk->indisk++;
        }
    }
#endif

skip:
    return ret;

}


int
skcache_entry_swapout_check(void *pp, void *vv)
{
	return _skcache_entry_swapout_check(pp, vv, 0);
}

int
skcache_entry_force_swapout_check(void *pp, void *vv)
{
	return _skcache_entry_swapout_check(pp, vv, 1);
}


#if 0
inline
static
int
_dicision(bucket_t *bk, skcache_entry_t *p, int force)
{
    time_t now;
    int elap, idle;
    int action;
	int chk;

    action = 0;

    now = time(NULL);

    if(p->issue_time > 0 ) {
        elap = now - p->issue_time;
    }
    else {
#if 1
        Error("_dc: empty issue_time");
#endif
        elap = -1;
    }

    if(p->hot > 0 ) {
        idle = now - p->hot;
    }
    else {
#if 0
        Trace("empty hot\n");
#endif
#if 0
        idle = MIN(skcache_mem_keep_cur/2, elap);
#endif
        idle = -1;
    }

#ifdef VS
    if(p->m_recorded) {
    Trace(" %p:: %8d %7d %c 0x%8x %3d, c %d, f %2d, idle %4d, elap %5d: ",
        p, p->oid, p->totallength,
        (p->ref > 0) ? 'R' : '-',
        p->m_recorded, p->d_recorded, p->cracked, p->count,
        idle, elap);
    }
    else {
    Trace(" %p:: %8d %7d %c %5s %8d, c %d, f %2d, idle %4d, elap %5d: ",
        p, p->oid, p->totallength,
        (p->ref > 0) ? 'R' : ' ',
        "-", p->d_recorded, p->cracked, p->count,
        idle, elap);
    }
#endif

#if 0
	Trace("lmd %d now %d diff %d ", p->lmd, now, now - p->lmd);
#endif

    /* action
        3   swapout + purge
        2   purge
        1   swapout
        0   no action
     */

    if(p->ref>0) {
#ifdef VS
Trace("refered ");
#endif
        action = 0;
    }
    else
	if(elap<0) {
#ifdef VS
Trace("junk    ");
#endif
        action = 2;
	}
	else
    if(p->cracked) {
#ifdef VS
Trace("cracked ");
#endif
        action = 2;
    }
    else
    if(p->d_recorded>=0) {
        if(p->m_recorded==NULL) {
#ifdef VS
Trace("already ");
#endif
            /* already swapouted */
            action = 0;
        }
        else {
#ifdef VS
Trace("mine-01 ");
#endif
			/* XXX */
            action = 1;
        }
    }
    else
    if(p->m_recorded==NULL) {
#ifdef VS
Trace("empty-b ");
#endif
        action = 2;
    }
    else
    if(p->totallength<=(off_t)0) {
#ifdef VS
Trace("empty-c ");
#endif
        action = 2;
    }
	else
    if(elap > skcache_entry_keep_cur) {
#ifdef VS
Trace("expire1 ");
#endif
        action = 2;
	}
	else
    if(p->expire>0 && p->expire <= now) {
#ifdef VS
Trace("expire2 ");
#endif
        action = 2;
    }
#if 0
    else
    if(p->lmd>0 && now - p->lmd > skcache_entry_keep_cur) {
#ifdef VS
Trace("expire3 ");
#endif
        action = 2;
	}
#endif
    else {

#if 1
		/*	
		 * XXX
		 * randomize forward swapout feature
	 	 */
		if(idle<=skcache_mem_keep_min) {
			/* nothing */
		}
		else {
			idle += rand() % skcache_mem_keep_cur;
		}
#endif

		if(force) {
			if(idle < 0) {
#ifdef VS
	Trace("neg,    ");
#endif
				action = 2;
			}
			else {
#ifdef VS
	Trace("shift   ");
#endif
				action = 3;
			}
		}
		else {
			if(idle < 0) {
#ifdef VS
	Trace("neg,    ");
#endif
				action = 2;
			}
			else
			if(idle >= skcache_mem_keep_cur) {
#ifdef VS
	Trace("shift   ");
#endif
				action = 3;
			}
			else {
#ifdef VS
	Trace("keep    ");
#endif
			}
		}
    }

skip:
    return action;

}
#endif

#if 0
int
_skcache_compaction(
		bucket_t *bk,
		skiplist_t *list, skiplist_cell_t **flist,
		int force)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x, *y, *n;
    int i;
    int ret;
    int diff;
    int chk;
	int action;
	int purge, swapout;
	int swapable = 0;

    ret = 0;

    x = list->header;
    for(i=0;i<=list->level;i++) {
            update[i] = x;
    }

    x = x->skip[0];

	MUTEX_LOCK(&stat_lock);
	if(swapout_token>0) {
		swapable = 1;
	}
	else {
		swapable = 0;
	}
	MUTEX_UNLOCK(&stat_lock);

    while(x!=list->term) {
        n = x->skip[0]; /* save next element */


		if(x->entity==NULL) {
			Error("NULL entity is found in skiplist of bk %p", bk);
			action = 0;
		}
		else {
			action 	= _dicision(bk, (skcache_entry_t*)x->entity, force);
		}
		swapout = action & 0x02;
		purge 	= action & 0x01;

		bk->op[action]++;

		if(swapout && swapable) {
			chk = skcache_entry_swapout(bk, (skcache_entry_t*)x->entity);
			if(chk) {
				/* fail swapout. don't purge */
				purge = 0;
			}
		}

        if(purge) {
delete:
			if(x->entity) {
				skcache_entry_purge(bk, (skcache_entry_t*)x->entity);
            	skcache_push_idle_entry(bk, (skcache_entry_t*)x->entity);
			}

            /*
             * skip current element
             * so, it is mean deletation of element
             * this routine don't care actual enitity procedure
             * (e.g, free(), unlink() or something )
             */
            for(i=0;i<=list->level;i++) {
                if(update[i]->skip[i]==x) {
                    update[i]->skip[i] = x->skip[i];
                }
            }

            /* add current element to free-list for recycle */
            y = x;
            y->entity = NULL;
            for(i=1;i<=list->maxlevel;i++) {
                y->skip[i] = NULL;
            }
            y->skip[0] = *flist;
            *flist = y;

            list->num--;

            ret++;
        }


no_delete:

        for(i=0;i<=list->level;i++) {
            if(update[i]->skip[i]==x)
                update[i] = x;
        }

        x = n;
    }

    return ret;
}

#endif


extern pthread_mutex_t stat_lock;
extern volatile int compaction_count;

#define GC_NTRY   (2)

int
skcache_compaction(bucket_t *bk, int v, int force, int e_or_f)
{
    int count;
    int deleted;
    int sum_deleted;
    skcache_entry_t *mark;
    int nelem_before, nelem_after;
    int nentry_before, nentry_after;
    int nfrag_before, nfrag_after;
    int hwater, lwater;
    double hrate, lrate;
    struct timeval t_before, t_after;
    struct timeval t_loopbefore, t_loopafter;
    int t_diff;
    time_t now;

    now = time(NULL);
#if 1
    Trace("skcache_compaction: %d bk# %d, %p #%d, %s %s\n",
        now, bk->bid, bk, v, force==1? "*FORCE*" : "NORMAL",
        e_or_f==XSK_ENTRY ? "ENTRY": "FRAG");

#endif

    mark = bk->entry_freelist;

    MUTEX_LOCK(&stat_lock);
    compaction_count++;
    MUTEX_UNLOCK(&stat_lock);


#if 0
    sk_rawdump(stderr, &bk->cup, skcache_label_ref);
#endif

    hwater = 9*bk->entry_all/10;
    lwater = 6*bk->entry_all/10;

    hrate  = 19.0/20.0;
    lrate  = 20.0/19.0;

#if 1
    /*
    Trace("    h/l water %d/%d, rate %4.2f/%4.2f bk->elemnum? %d\n",
        hwater, lwater, hrate, lrate,
        bk->elemnum
        );
    */

    Trace("  keeptime mem [%d..%d) disk [%d..%d] \n",
        0, skcache_mem_keep_cur,
        skcache_mem_keep_cur, skcache_entry_keep_cur);

    Trace("  elem  %4d/%-5d // %p\n",
        bk->cup.num, bk->elemnum, bk->elem_freelist);

    Trace("  entry %4d/%-5d // hh %4d h %4d l %4d : %p\n",
        bk->entry_use, bk->entry_all,
        bk->entry_hhwater, bk->entry_hwater, bk->entry_lwater,
        bk->entry_freelist);

    Trace("  frag  %4d/%-5d // hh %4d h %4d l %4d : %p\n",
        bk->frag_use, bk->frag_all,
        bk->frag_hhwater,  bk->frag_hwater, bk->frag_lwater,
        bk->frag_freelist);

    Trace("  - uv %p skiplist %4d/%d, entry %4d/%d, frag %4d/%d\n",
        bk,
        bk->cup.num, bk->elemnum,
        bk->entry_use, bk->entry_all,
        bk->frag_use,  bk->frag_all);
#endif

#if 0
    Log("; GC bk %p %d %s %d %d",
        bk, v, force==1? "*force*" : "normal",
        skcache_mem_keep_cur, skcache_entry_keep_cur);
#endif

    gettimeofday(&t_loopbefore, NULL);

    sum_deleted = 0L;
    count = 0;
    do {
        nelem_before    = bk->cup.num;
        nentry_before   = bk->entry_use;
        nfrag_before    = bk->frag_use;

        bk->inmemory = 0;
        bk->indisk   = 0;
		bk->op[0] = bk->op[1] = bk->op[2] = bk->op[3] = 0;

        gettimeofday(&t_before, NULL);

 assert(bk->cup.num==bk->entry_use);

#if 1
        if(force==1) {
            deleted = Zsk_mapdelete(&bk->cup, &bk->elem_freelist,
                skcache_entry_force_swapout_check, (void*)bk);
 assert(bk->cup.num==bk->entry_use);
        }
        else {
            deleted = Zsk_mapdelete(&bk->cup, &bk->elem_freelist,
                skcache_entry_swapout_check, (void*)bk);
 assert(bk->cup.num==bk->entry_use);
        }
#endif
#if 0
		deleted =
			_skcache_compaction(bk, &bk->cup, &bk->elem_freelist, force);
 assert(bk->cup.num==bk->entry_use);
#endif

        gettimeofday(&t_after, NULL);

        nelem_after     = bk->cup.num;
        nentry_after    = bk->entry_use;
        nfrag_after     = bk->frag_use;

        t_diff = KOTETU_SUB_TIME(t_after, t_before);

#if 1
        Trace("  %d uv %p skiplist %4d/%d, entry %4d/%d, frag %4d/%d\n",
            count,
            bk,
            bk->cup.num, bk->elemnum,
            bk->entry_use, bk->entry_all,
            bk->frag_use,  bk->frag_all);
        Trace("    deleted %d, rest memory %4d disk %4d\n",
            deleted, bk->inmemory, bk->indisk);
		Trace("    keep %4d, swapout %4d, purge %4d, swapout+purge %4d <%d>\n",
			bk->op[0], bk->op[1], bk->op[2], bk->op[3],
			bk->op[0] + bk->op[1] + bk->op[2] + bk->op[3]
			);
#endif

#if 0
        Trace("    elem:  before %4d after  %4d diff %4d\n",
            nelem_before, nelem_after, nelem_before - nelem_after);
        Trace("    entry: before %4d after  %4d diff %4d\n",
            nentry_before, nentry_after, nentry_before - nentry_after);
        Trace("    frag:  before %4d after  %4d diff %4d\n",
            nfrag_before, nfrag_after, nfrag_before - nfrag_after);
#endif
        if(nelem_before - nelem_after != deleted ) {
            Error("  change of list elements do not match deleted");
        }
        if(nentry_before - nentry_after != deleted ) {
            Error("  change of entries do not match deleted");
        }


        MUTEX_LOCK(&config_lock);
        /* entry */
        if(e_or_f==XSK_ENTRY) {
#if 0
            if(nentry_before >= bk->entry_hwater) {
                Trace("decrease entry 1\n");
                skcache_entry_keep_cur =
                    MAX((int)(hrate*skcache_entry_keep_cur),
                        skcache_entry_keep_min);
            }
            else
            if(nentry_after >= bk->entry_hwater) {
                Trace("decrease entry 2\n");
                skcache_entry_keep_cur =
                    MAX((int)(hrate*skcache_entry_keep_cur),
                        skcache_entry_keep_min);
            }
            else
            if(nentry_after <= bk->entry_lwater) {
                Trace("increase entry 1\n");
                skcache_entry_keep_cur =
                    MIN((int)(lrate*skcache_entry_keep_cur),
                        skcache_entry_keep_max);
            }
#endif

#if 0
            Trace("increase mem 1\n");
                skcache_mem_keep_cur =
                    MIN((int)(lrate*skcache_mem_keep_cur),
                        skcache_mem_keep_max);
#endif
        }
        /* frag */
		else
        if(e_or_f==XSK_FRAG) {

#if 0
            if(bk->frag_use >= bk->frag_hwater) {
                Trace("decrease mem 2\n");
#if 0
                skcache_mem_keep_cur =
                    MAX((int)(hrate*skcache_mem_keep_cur),
                        skcache_mem_keep_min);
#endif
                skcache_mem_keep_cur =
					MAX(skcache_mem_keep_cur-10, skcache_mem_keep_min);
            }
            else
            if(bk->frag_use <= bk->frag_lwater) {
                Trace("increase mem 2\n");
#if 0
                skcache_mem_keep_cur =
                    MIN((int)(lrate*skcache_mem_keep_cur),
                        skcache_mem_keep_max);
#endif
                skcache_mem_keep_cur =
                    MIN(skcache_mem_keep_cur+10, skcache_mem_keep_max);
            }
#endif

        }
        else {
		}

#if 0
        skcache_mem_keep_cur =
            MIN(skcache_mem_keep_cur, skcache_entry_keep_cur/2);
#endif
#if 1
        skcache_mem_keep_cur =
            MIN(skcache_mem_keep_cur, skcache_entry_keep_cur);
#endif

        MUTEX_UNLOCK(&config_lock);

#if 1
        Trace("  keeptime mem [%d..%d) disk [%d..%d] \n",
            0, skcache_mem_keep_cur,
            skcache_mem_keep_cur, skcache_entry_keep_cur);
#endif

        sum_deleted += deleted;
        count++;


 assert(check_bkentry(bk, "GC-"));

    } while(count<GC_NTRY && nentry_before == nentry_after);

    gettimeofday(&t_loopafter, NULL);

    t_diff = KOTETU_SUB_TIME(t_loopafter, t_loopbefore);

    Trace("  take %d msec ave. %.2f , bk# %d, %p\n",
		t_diff, (double)t_diff/count, bk->bid, bk);
    sim_stat_add_value(&compaction_time, t_diff/100);


#if 0
    sk_rawdump(stderr, &bk->cup, skcache_label_ref);
#endif

#if 0
    Trace("  ret %d\n", c);
#endif
	bk->last_gc = now;
	bk->gc_push = 0;

 assert(check_bkentry(bk, "GCn"));

    return sum_deleted;
}



int
skcache_check_mem(int d1)
{
    struct timeval begin, end;
    time_t reftime;
    bucket_t *bk;
    int chk;
	int xfrag_hwater, xentry_hwater;

    gettimeofday(&begin, NULL);
    reftime = begin.tv_sec;

#if 0
Trace("check_mem: d1 %4d\n", d1);
#endif

	if(d1<0||d1>=RADIX_H1_MAX) {
		Error("ignore bucket position %d",  d1);
		goto out;
	}
	if(d1>=radix_h1) {
		Trace("check_mem: out of current bucket range <%dd>, no change\n", d1);
		goto out;
	}	

    bk = &skcache_bucket_base[d1];

#ifdef VM
Trace("check_mem: %p [%04d] %4d th %08x %d\n",
    bk, d1, radix_h1, reftime);
#endif

	LOCK_BK(bk);
 assert(bk->cup.num==bk->entry_use);

    /* NOTE: this routine use 'hwater', not 'hhwater'
     * hwater is lower than hhwater.  Using hwater, this routine
     * would compact before ondemand compaction by hhwater.
     * It makes fast response for clients.
     */
	/* randomize */
#if 1
	xfrag_hwater  = bk->frag_hwater  - (rand()% bk->frag_all/10);
	xentry_hwater = bk->entry_hwater - (rand()% bk->entry_all/10);
#else
	xfrag_hwater  = bk->frag_hwater;
	xentry_hwater = bk->entry_hwater;
#endif

    if(bk->frag_use >= xfrag_hwater || bk->frag_freelist==NULL) {
        Trace("check_mem: frag compaction bk# %d %p\n", d1, bk);
        chk = skcache_compaction(bk, -97, 0, 1);
        Trace("compaction func. return %d <8>\n", chk);
        if(chk==0 &&
    		(bk->frag_use >= xfrag_hwater || bk->frag_freelist==NULL)) {
            chk = skcache_compaction(bk, -45, 1, 1);
            Trace("compaction func. return %d <13>\n", chk);
        }
    }
	else
    if(bk->entry_use >= xentry_hwater || bk->entry_freelist==NULL) {
        Trace("check_mem: entry compaction bk# %d\n", d1);
        chk = skcache_compaction(bk, -123, 0, 0);
        Trace("compaction func. return %d <9>\n", chk);
        if(chk==0) {
            chk = skcache_compaction(bk, -73, 1, 0);
            Trace("compaction func. return %d <12>\n", chk);
        }
    }

 assert(bk->cup.num==bk->entry_use);
    UNLOCK_BK(bk);


out:

    gettimeofday(&end, NULL);
    sim_stat_add_value(&checkmem_time, KOTETU_SUB_TIME(end, begin)/100);

    return 0;
}

typedef struct {
	int no;
	int act;
	time_t date;
	off_t sz;
} tup;

int
comp_tup(const void* xp, const void* yp)
{
	tup *x = (tup*)xp;
	tup *y = (tup*)yp;

	if(x->act == y->act) {
		if(x->date == y->date) {
			return y->sz - x->sz;			/* y-x greater */
		}
		else {
			return x->date - y ->date;		/* x-y smaller */
		}
	}
	else {
		return y->act - x->act;				/* y-x greater */
	}
}

int
comp_tup_time(void*xp,void*yp)
{
	tup *x = (tup*)xp;
	tup *y = (tup*)yp;
	
	return x->date - y->date;
}

#define TUP_MAX	(512)

#define H4_DIR_COUNT_MAX	(128)
#define H4_DIR_SIZE_MAX		(128*K(10))
/*
#define H4_DIR_SIZE_MAX		(K(52))
*/

volatile int dir_count_max = H4_DIR_COUNT_MAX;
volatile off_t dir_size_max =  H4_DIR_SIZE_MAX;


int
skcache_check_dir(int d1, int d2, int d3, int force)
{
    char basename[L_PATH];
    char dirname[L_PATH];
    char dir2name[L_PATH];
    char tmp[L_PATH];
    char *namechk;
    dirent_t    *elem;
    DIR         *dirp;
#ifdef PC_NAME_MAX
    dirent_t ent[PC_NAME_MAX+1];
#else
#ifdef _PC_NAME_MAX
    dirent_t ent[_PC_NAME_MAX+1];
#else
    dirent_t ent[128];
#endif
#endif
    char name[L_PATH];
/*
    struct stat sbuf;
*/
    int dfd;
    char buf[L_BUF];
    int n;
	int i;
    int nr;
    int chk;
    off_t sum;
	off_t fsize;
    int lcount, dcount, ecount;
    off_t lsize, dsize, esize;
    int strange = 0;
    int action;
    int life;
    time_t reftime;
    time_t now;

	tup rec[TUP_MAX];

    struct timeval begin, end;

    bucket_t *bk;
    skcache_entry_t en;
	int c;

    int sig = 0L;
    int sig_mask = (1L << 31) ;
    int sig_pos = 0;

    gettimeofday(&begin, NULL);
    reftime = begin.tv_sec;

#if 0
Trace("check_dir: d1,d2,d3 %4d,%4d,%4d time %d force? %d\n",
   d1, d2, d3, reftime, force);
#endif

	if(d1<0||d1>=RADIX_H1_MAX
		|| d2<0||d2>=RADIX_H2_MAX
			|| d3<0||d3>=RADIX_H3_MAX) {
		Error("ignore directory position %d,%d,%d",  d1, d2, d3);
		goto out;
	}

	if(d1>=radix_h1 || d2>=radix_h2 || d3>=radix_h3) {
		int rm_fr(char*);
		if(force) {
#if 1
			Trace("check_dir: out of current directory range <%d,%d,%d>, try clear\n",
				d1, d2, d3);
#endif

#if 1
			namechk = assign_dirname_bynumber(d1, d2, -1, dirname);
			if(namechk) {
#if 1
				Trace("check_dir: clear dir. recursive %s\n", dirname);
#endif
				rm_fr(dirname);
			}
#endif
#if 0
			namechk = assign_dirname_bynumber(d1, d2, d3, dirname);
			if(namechk) {
				rm_fr(dirname);
			}
#endif
		}
		else {
#if 1
			Trace("check_dir: out of current directory range <%d,%d,%d>, no change\n",
				d1, d2, d3);
#endif
		}

		goto out;
	}	


    bk = &skcache_bucket_base[d1];

#ifdef VD
Trace("check_dir: %p [%03d:%03d:%03d] th %08x %d %s\n",
    bk, d1, d2, d3, pthread_self(), reftime, force ? "FORCE" : "normal");
#endif
#if 0
Trace("    %4d/%4d/%4d\n", radix_h1, radix_h2, radix_h3);
#endif

    LOCK_BK(bk);
 assert(bk->cup.num==bk->entry_use);
 assert(check_bkentry(bk, "cd0"));


    if(bk->dirty2 >= 0 || bk->dirty3 >= 0) {
        Error(
"check_dir: %p [%03d:%03d:%03d], [%03d:%03d:%03d] is used by another thread",
            bk, d1, d2, d3, d1, bk->dirty2, bk->dirty3);
        UNLOCK_BK(bk);
        goto out;
    }

    bk->dirty2 = d2;
    bk->dirty3 = d3;

    UNLOCK_BK(bk);


disk_operations:


    n = 0;
    sum = (off_t)0L;

    namechk = assign_dirname_bynumber(d1, d2, d3, dirname);

#if 0
    Trace(" dir '%s'\n", dirname);
#endif


	/***
	 *** scan directories
	 ***/

	c = 0;
try_opendir:
	c++;
    dirp = opendir(dirname);
    if(dirp==NULL) {
        if(errno==ENOENT && c<DISK_RETRY) {
			if(force) {
#if 0
				Trace("setup new dir %d,%d,%d\n", d1, d2, d3);
#endif
				Trace("not found dir '%s', build that.\n", dirname);
				build_dir_bynumber(d1, d2, d3);
			}

            goto fail_opendir;
        }
		if(errno==EINTR && c<DISK_RETRY) {
			goto try_opendir;
		}

        Error("fail opendir() '%s' %d (%d)", dirname, c, errno);

        goto fail_opendir;
    }

	i = 0;
    while(1) {
        chk = readdir_r(dirp, ent, &elem);
        if(elem==NULL) {
#if 0
			Trace("dir '%s' is empty\n", dirname);
#endif
            break;
        }

        if(elem->d_name[0]=='.' && elem->d_name[1]=='\0') {
            continue;
        }
        if(elem->d_name[0]=='.'
            && elem->d_name[1]=='.' && elem->d_name[2]=='\0') {
            continue;
        }
		
		i++;

        strcpy(name, dirname);
        strcat(name, "/");
        strcat(name, elem->d_name);

#if 0
		Trace("    %4d %-14s %s\n", i, elem->d_name, name);
#endif


        life = -1;
        action = 0;

#if 0
        chk = stat(name, &sbuf);
        if(chk) {
            Trace(" f %4d '%s' error (%d)\n", n, name, errno);
            strange++;
            continue;
        }

        sum += (off_t) sbuf.st_size;
#endif


		c = 0;
try_open:
		c++;
        dfd = open(name, O_RDONLY);
        if(dfd<0) {
			if(errno==EINTR && c<DISK_RETRY) {
				goto try_open;
			}
			Error("check_dir: fail open() '%s' %d (%d)\n",
				name, c, errno);
			action = 1;
			goto next;
		}

		c = 0;
try_lseek:
		c++;
		fsize = lseek(dfd, 0, SEEK_END);
		if(fsize<=(off_t)0) {
			if(errno==EINTR && c<DISK_RETRY) {
				goto try_lseek;
			}
			Error("check_dir: fail lseek() '%s' %d (%d)\n",
				name, c, errno);
			action = 1;
			goto done_read_cntl;
		}
		if(fsize>(off_t)0) {
			lseek(dfd, 0, SEEK_SET);
		}
		
		sum += fsize;

		c = 0;
try_read_cntl:
		c++;
		nr = read(dfd, &en, sizeof(skcache_entry_t));
		if(nr<0) {
			if(errno==EINTR && c<DISK_RETRY) {
				goto try_read_cntl;
			}
			Error("check_dir: file# %d, fail read() '%s' %d (%d)\n",
				n, name, c, errno);
			action = 1;
			goto done_read_cntl;
		}

		if(nr<sizeof(skcache_entry_t)) {
			action = 2;
#if 1
			Error("check_dir: ignore file# %d, too short %d/%d",
				n, nr, sizeof(skcache_entry_t));
#endif
			en.issue_time = -1;
			goto done_read_cntl;
		}

		if(en.magic[0]!=MAGIC_0 || en.magic[1]!=MAGIC_1) {
			action = 2;
#if 1
Error("check_dir: ignore file# %d, wrong magic %02x %02x / %02x %02x '%s'",
		n, en.magic[0], en.magic[1], MAGIC_0, MAGIC_1, name);
#endif
			en.issue_time = -1;
			goto done_read_cntl;
		}

{
	int xh1, xh2, xh3, xh4;

		calc_hash(en.name, &xh1, &xh2, &xh3, &xh4);

		if(en.h1!=xh1 || en.h2!=xh2 || en.h3!=xh3 || en.h4!=xh4) {
			action = 2;
#if 1
Trace("\
checkdir: file '%s'\n\
	enN %03d:%03d:%03d:%03d %s\n\
	xhN %03d:%03d:%03d:%03d %s\n",
	name,
	en.h1, en.h2, en.h3, en.h4, en.name,
	xh1, xh2, xh3, xh4, en.name);
#endif
#if 0
Trace("check_dir: ignore hash, stored %03d:%03d:%03d:%03d, we guess %03d:%03d:%03d:%03d\n",
		en.h1, en.h2, en.h3, en.h4, xh1, xh2, xh3, xh4);
#endif
			en.issue_time = -1;
			goto done_read_cntl;
		}

}

		if(en.issue_time<=0) {
			action = 1;
		}
		else {
#if 1
			/*XXX*/
			if(en.issue_time<min_date) {
				action = 2;
			}
			else
			if(en.issue_time>reftime) {
				action = 2;
			}
			else
#endif
			if(reftime - en.issue_time > skcache_entry_keep_cur) {
				action = 1;
			}
		}

		if(!action) {
			if(en.lmd<=0) {
				action = 1;
			}
#if 0
			if(en.lmd>0 && reftime - en.lmd > skcache_entry_keep_cur) {
				action = 1;
			}
#endif

			if(en.expire>0 && reftime - en.expire > 0) {
				action = 1;
			}
		}


done_read_cntl:
		(void) 1;

read_done:

		close(dfd);


next:

#ifdef VD2
        Trace(" f %4d %4d %4d %d %d %8d %9ld %s\n",
            i, n, atoi(elem->d_name), action,
			en.issue_time, reftime - en.issue_time, (long)fsize, name);
#endif

		if(n>=TUP_MAX) {
			Error("tup overflow %d/%d", n, TUP_MAX);
		}
		else {
			rec[n].no = atoi(elem->d_name);
			rec[n].act = action;
			rec[n].date = en.issue_time;
			rec[n].sz = fsize;
			n++;
		}
	}

    closedir(dirp);



	/***
	 *** main body of compaction
	 ***/

	qsort(rec, n, sizeof(tup), comp_tup);

    lcount 	= 0;
    dcount 	= 0;
    lsize 	= 0;
    dsize 	= 0;
    ecount 	= 0;
    esize 	= 0;

#if 0
	Trace("sum %lld, aim %lld\n",
		(long long)sum, (long long)dir_size_max);
#endif

	for(i=0;i<n;i++) {
		sprintf(name, "%s/%d", dirname, rec[i].no);

#if 0
		Trace("%4d: no# %5d act %2d date %d size %9ld\n",
			i, rec[i].no, rec[i].act, rec[i].date, rec[i].sz);
#endif

		if(rec[i].act==0) {
			if(sum - esize - dsize > dir_size_max) {
#if 0
				Trace("\t\tremove 1\n");
#endif
				rec[i].act = 1;
			}
			else {
				if(rec[i].sz > dir_size_max) {
#if 0
					Trace("\t\tremove 2\n");
#endif
					rec[i].act = 1;
				}
				else {
#if 0
					Trace("\t\tkeep\n");
#endif
				}
			}
		}

        switch(rec[i].act) {
		case 0:
            sig |= (rec[i].no+1);

            lcount++;
            lsize += rec[i].sz;
			break;

        case 1:

			c = 0;
try_unlink1:
			c++;
            chk = unlink(name);
			if(chk) {
				Error("check_dir: 1 - fail unlink() %s %d (%d)",
					name, c, errno);
			}

            dcount++;
            dsize += rec[i].sz;
            break;

        case 2:
        default:

			c = 0;
try_unlink2:
			c++;
            chk = unlink(name);
			if(chk) {
				Error("check_dir: 2 - fail unlink() %s %d (%d)",
					name, c, errno);
			}

            ecount++;
            esize += rec[i].sz;
            break;

		}

#if 0
		Trace("\tl %9ld d %9ld e %9ld rest %9ld\n",
			lsize, dsize, esize, sum - esize - dsize);
#endif

    }



    now = time(NULL);

    sig_pos = d1*radix_h2*radix_h3 + d2*radix_h3 + d3;


#ifdef VD
    Trace(" dd %6d %-24s %4d %8x %7ld + %d\
/ live %4d %7ld / dead %4d %7ld\
/ old %4d %7ld / %2d\n",
        sig_pos, dirname, n, sig, (long)sum, strange,
        lcount, (long)lsize, dcount, (long)dsize, ecount, (long)esize,
        now - reftime
        );
#endif
	dir_sz[sig_pos] = lsize;

#if 0
	if(lsize != (off_t)0L || dsize != (off_t)0L) {
    	Trace(" dd %6d %-24s %4d %8x %7ld + %d\
/ live %4d %7ld / dead %4d %7ld\
/ old %4d %7ld / %2d\n",
        sig_pos, dirname, n, sig, (long)sum, strange,
        lcount, (long)lsize, dcount, (long)dsize, ecount, (long)esize,
        now - reftime
        );
	}
#endif

    MUTEX_LOCK(&stat_lock);

	if(dcount>0 && dsize>0) {
		swap_count  -= dcount;
		swap_size   -= dsize;

		if(swap_count<=0LL || swap_size<=0LL) {
			Trace("swap count or size is negative, count %lld size %lld\n",
				swap_count, swap_size);
		}
	}

	if(force) {
		dir_sig[sig_pos] = sig;
	}

    MUTEX_UNLOCK(&stat_lock);


fail_opendir:

	if(bk) {
		LOCK_BK(bk);
 assert(check_bkentry(bk, "cd8"));

		bk->dirty2 = -1;
		bk->dirty3 = -1;
 assert(check_bkentry(bk, "cd9"));

		UNLOCK_BK(bk);
	}

out:

    gettimeofday(&end, NULL);
    sim_stat_add_value(&checkdir_time, KOTETU_SUB_TIME(end, begin)/100);

    return 0;
}


volatile int sweeper_seq    = 0;
volatile int sweeper_count  = 0;
volatile static int sweeper_pos=0;
static int sweeper_phase = 0;
int full_sweep_mode = 0;

/* how many tics are required to work sweeper */
#ifndef SWEEP_CYCLE
#define SWEEP_CYCLE (1)
#endif
#ifndef SWEEP_BOOT_CYCLE
#define SWEEP_BOOT_CYCLE (1)
#endif

/* how many direcotry are sweeped per chance */
#ifndef SWEEP_NTRY
#define SWEEP_NTRY  (1)
#endif


volatile int sweep_boot_cycle = SWEEP_BOOT_CYCLE;
volatile int sweep_cycle = SWEEP_CYCLE;


int
skcache_dir_sweeper(void)
{
    int f, i;
    int d1, d2, d3;
    int v;
    int chk;
    int p;
	int nd1;

	int dir_max=RADIX_H1_MAX*RADIX_H2_MAX*RADIX_H3_MAX;

    MUTEX_LOCK(&stat_lock);
    p = sweeper_phase;
    f = sweeper_seq;
    sweeper_seq =  (sweeper_seq+1) % 1000000;
    MUTEX_UNLOCK(&stat_lock);

	if(p==1) {
		goto phase_one;
	}

phase_zero:
	if(f%sweep_boot_cycle!=0) {
		return 0;
	}

#if 0
	Trace("sweep 0: th %08x f %9d -> %4d\n",
		pthread_self(), f, f%sweep_boot_cycle);
#endif

	MUTEX_LOCK(&stat_lock);
	sweeper_count++;
	v = sweeper_pos;

	if(full_sweep_mode) {
		sweeper_pos = (sweeper_pos + 1 ) % dir_max;

		d3 = (v/(RADIX_H2_MAX*RADIX_H1_MAX))%RADIX_H3_MAX;
		d2 = (v%(RADIX_H2_MAX*RADIX_H1_MAX))/RADIX_H1_MAX;
		d1 = (v%(RADIX_H2_MAX*RADIX_H1_MAX))%RADIX_H1_MAX;

	}
	else {
        sweeper_pos = (sweeper_pos + 1 ) % dir_num;

        d3 = (v/(radix_h2*radix_h1))%radix_h3;
        d2 = (v%(radix_h2*radix_h1))/radix_h1;
        d1 = (v%(radix_h2*radix_h1))%radix_h1;
	}
	MUTEX_UNLOCK(&stat_lock);

#if 0
	Trace("  p0 th %08x %8d -- %6d -> [%03d:%03d:%03d] %s\n",
		pthread_self(), f, v, d1, d2, d3, full_sweep_mode ? "FULL" : "normal");
#endif

	if(full_sweep_mode) {
        int rm_fr(char*);
		char dirname[L_PATH];
		char *namechk;

    	if(d1>=radix_h1 || d2>=radix_h2 || d3>=radix_h3) {
            namechk = assign_dirname_bynumber(d1, d2, -1, dirname);
            if(namechk) {
#if 1
                Trace("sweeper: clear dir %03d:%03d:%03d %s\n", d1, d2, d3, dirname);
#endif
                rm_fr(dirname);
            }
        }
		else {
			chk = skcache_check_dir(d1, d2, d3, 0);
		}
	}
	else {
		chk = skcache_check_dir(d1, d2, d3, 1);
	}


	MUTEX_LOCK(&stat_lock);
	if(sweeper_pos == 0) {
		sweeper_phase = 1;
	}
	MUTEX_UNLOCK(&stat_lock);

	goto done;

phase_one:
    if(f%sweep_cycle!=0) {
        return 0;
    }
#if 0
	Trace("sweep 1: th %08x f %9d -> %4d\n",
		pthread_self(), f, f%sweep_cycle);
#endif

    for(i=0;i<SWEEP_NTRY;i++) {
        MUTEX_LOCK(&stat_lock);
        if(sweeper_count%dir_num==0) {
            Trace("first dir\n");
        }
        sweeper_count++;
        v = sweeper_pos;
        sweeper_pos = (sweeper_pos + 1 ) % dir_num;
        MUTEX_UNLOCK(&stat_lock);

        if(f<0) {
            f = v;
        }
        d3 = (v/(radix_h2*radix_h1))%radix_h3;
        d2 = (v%(radix_h2*radix_h1))/radix_h1;
        d1 = (v%(radix_h2*radix_h1))%radix_h1;
#if 0
        Trace("  p1 th %08x %8d %2d %6d -> [%03d:%03d:%03d]\n",
			pthread_self(), f, i, v, d1, d2, d3);
#endif
        chk = skcache_check_dir(d1, d2, d3, 1);

		/* previous bucket */
#if 0
		/* neighbor is often busy */
		nd1 = (d1 + radix_h1 - 1) % radix_h1;
        chk = skcache_check_mem(nd1);
#endif

#if 0
		nd1 = (d1 + radix_h1 + radix_h1/2 ) % radix_h1;
        chk = skcache_check_mem(nd1);
#endif

#if 0
        if(v%(dir_num/4)==0) {

 			/* XXX */
            sim_stat_show_lock(&swapload_time,  stderr);   
            sim_stat_show_lock(&swapin_time,    stderr);
            sim_stat_show_lock(&swapout_time,   stderr);
            sim_stat_show_lock(&compaction_time, stderr);
            sim_stat_show_lock(&checkdir_time,  stderr);
            sim_stat_show_lock(&checkmem_time,  stderr);

        }
#endif

    }

done:

    return 0;
}





int
skcache_calc_param2(long long want,
    int *xbk_num, int *xsk_level, int *xffl_size, double r,
    int *xsk_len, int *xffl_num)
{
    long long bksize;
    long a, b, c;

#if 0
    Trace("skcache_calc_param: want %lld\n", want);
#endif

#if 0
    Trace("    specified xbk_num %d, xsk_level %d, xffl_size %d r %.2f\n",
        *xbk_num, *xsk_level, *xffl_size, r);
#endif

    if(*xffl_size<0) {
        *xffl_size = 1460;
    }

    bksize = want/(*xbk_num) - sizeof(skcache_entry_t);

    a = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(*xsk_level+1);
    b = sizeof(skcache_entry_t);
    c = sizeof(skcache_frag_t)+ *xffl_size;

#if 0
    Trace("    bucket size %lld\n", bksize);
    Trace("    a %ld\n", a);
    Trace("    b %ld\n", b);
    Trace("    c %ld\n", c);
#endif

    *xsk_len = (long)
        (bksize - b) / (a + b + (long)(r*c));
    *xffl_num = (long)(r * *xsk_len);

#if 0
    Trace("    solved    xsk_len %d, xffl_num %d\n",
        *xsk_len, *xffl_num);
#endif


    return *xsk_len;
}


long long
skcache_calc_size(
    int *xbk_num, int *xsk_level, int *xsk_len, int *xffl_len, int *xffl_size)
{
    long long r;
    int esk;
    int tsk;
    int ten;
    int tfl;
    long long ebk;
    long long tbk;

#if 1
    Trace("skcache_calc_size:\n");
    Trace("  bucket:\n");
#endif

    r = (((sizeof(void*)+sizeof(skiplist_cell_t**)+sizeof(skiplist_cell_t*)*(*xsk_level+1)) + sizeof(skcache_entry_t))* (*xsk_len)
        + sizeof(skcache_entry_t)
        + (sizeof(skcache_frag_t) + *xffl_size) * (*xffl_len))*(*xbk_num);

    esk = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(*xsk_level+1);
    tsk = esk * *xsk_len;
#if 1
    Trace("    skiplist:\n");
    Trace("      level %d\n", *xsk_level);
    Trace("      -> element size of skiplist %ld\n", esk);
    Trace("      length of skiplist %d\n", *xsk_len);
#endif

    ten = sizeof(skcache_entry_t)*(*xsk_len+1);
    tfl = (sizeof(skcache_frag_t)+ *xffl_size)*(*xffl_len);
    ebk = (long long)(sizeof(bucket_t)+tsk+ten+tfl);
#if 1
    Trace("    -> size of skiplist            %12ld %4ldM+%07ld %7.2f%%\n",
        tsk, (long)dM(tsk), (long)mM(tsk), ((double)tsk*100)/ebk);
    Trace("    size of free-list for entities %12ld %4ldM+%07ld %7.2f%%\n",
        ten, (long)dM(ten), (long)mM(ten), ((double)ten*100)/ebk);
    Trace("    size of fragment free-list     %12ld %4ldM+%07ld %7.2f%%\n",
        tfl, (long)dM(tfl), (long)mM(tfl), ((double)tfl*100)/ebk);
    Trace("  -> size of bucket                %12lld %4ldM+%07ld\n",
        ebk, (long)dM(ebk), (long)mM(ebk));
#endif


    tbk = ebk * *xbk_num;

#if 1
    Trace("  number of bucket %ld\n", *xbk_num);
    Trace("-> total size of all buckets       %12lld %4ldM+%07ld\n",
        tbk, (long)dM(tbk), (long)mM(tbk));
    Trace("-> total number of entries         %12ld\n", *xsk_len * *xbk_num);
    Trace("-> total number of frags           %12ld\n", *xffl_len * *xbk_num);
#endif

#if 0
    Trace("@ %4d %2d %5d -> %9d %9d\n",
        *xbk_num, *xsk_level, *xffl_len,
        *xsk_len * *xbk_num,  *xffl_len * *xbk_num);
#endif

    return r;
}


skcache_frag_t*
skcache_pull_new_frag(bucket_t *bk)
{
    skcache_frag_t* cur;
    int chk;

 assert(bk->cup.num==bk->entry_use);

#if 0
Trace("new_frag bk %p, used %d/%d\n", bk, bk->frag_use, bk->frag_all);
#endif

#ifndef NDEBUG
  {
    int c;

    skcache_frag_t* cur;
    c = 0;
    cur = bk->frag_freelist;
    while(cur) {
        c++;
        if(cur->next==cur) {
            Error("new_frag: found frag loop bk %p, %d th frag %p",
                bk, c, cur);
        }
        cur = cur->next;
    }

    if(bk->frag_use + c == bk->frag_all) {
        /* nothing */
    }
    else {
        Error("new_frag: bk %p, frag use/all %d/%d -> %d, rest %d, diff %d",
            bk, bk->frag_use, bk->frag_all,
            bk->frag_all - bk->frag_use,
            c, (bk->frag_all - bk->frag_use) - c
            );
    }

  }
#endif

#if 0
    if(bk->frag_use >= bk->frag_hhwater || bk->frag_freelist==NULL) {
#if 1
        Trace("new_frag bk %p, used %4d/%d, p %p\n",
            bk, bk->frag_use, bk->frag_all, bk->frag_freelist);
#endif
        chk = skcache_compaction(bk, -7, 0, 1);
        Trace("compaction func. return %d <4a>\n", chk);
 assert(bk->cup.num==bk->entry_use);
		if(chk==0) {
        	chk = skcache_compaction(bk, -127, 1, 1);
        Trace("compaction func. return %d <4a>\n", chk);
 assert(bk->cup.num==bk->entry_use);
		}
    }
#endif

    if(bk->frag_freelist==NULL) {
        Error("empty frag free list, bk# %d %p <4v>", bk->bid, bk);

        return NULL;
    }
    if(bk->frag_use >= bk->frag_all) {
        Error("full  frag list, bk# %d %p <4u>", bk->bid, bk);

        return NULL;
    }
#if 1
    if(bk->frag_use==bk->frag_all-1) {
        Trace("skcache_pull_new_frag: last frag of bk# %d %p\n", bk->bid, bk);
    }
#endif

#if 0
    if(bk->frag_all - bk->frag_use <5) {
        skcache_frag_t *p;
        int i;
        Trace("skcache_pull_new_frag: bk# %d %p\n", bk->pid, bk);
        p = bk->frag_freelist;
        i = 0;
        while(p) {
            Trace("    %2d: %p next %p\n", i, p, p->next);
            if(p->next==p) {
                Trace("    loop\n");
                break;
            }
            p = p->next;
            i++;
        }
    }
#endif

    cur = bk->frag_freelist;
    bk->frag_freelist = cur->next;
    bk->frag_use++;
    cur->next = NULL;
    cur->len  = 0;

    return cur;
}

int
skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf)
{
#if 1
    if(bk->frag_use<=0) {
        Error("skcache_push_idle_frag: underflow bk %p, frag %p", bk, xf);
    }
#endif

#if 0
    if(bk->frag_all - bk->frag_use <5) {
        skcache_frag_t *p;
        int i;
        Trace("skcache_push_idle_frag: bk %p, frag %p\n", bk, xf);
        p = bk->frag_freelist;
        i = 0;
        while(p) {
            Trace("    %2d: %p next %p\n", i, p, p->next);
            if(p->next==p) {
                Trace("    loop\n");
                break;
            }
            p = p->next;
            i++;
        }
    }
#endif

    xf->next = bk->frag_freelist;
    bk->frag_freelist = xf;
    bk->frag_use--;

    return 0;
}



int check_bkinfo();



skcache_entry_t*
skcache_pull_new_entry(bucket_t *bk)
{
    int chk;
    skcache_entry_t* newone;

#if 0
Trace("new_entry bk %p, used %d/%d\n", bk, bk->entry_use, bk->entry_all);
#endif

 assert(bk->cup.num==bk->entry_use);

	/***
	 *** on-demand memory compaction
	 ***/
#if 0
    if(bk->entry_use >= bk->entry_hhwater || bk->entry_freelist==NULL
        || bk->elem_freelist == NULL) {

#if 1
        Trace("new_entry bk %p, used %d/%d, p %p\n",
            bk, bk->entry_use, bk->entry_all, bk->entry_freelist);
#endif
        chk = skcache_compaction(bk, bk->bid 0, 0);
        Trace("compaction func. return %d <1>\n", chk);
 assert(bk->cup.num==bk->entry_use);

        if(chk==0) {
            chk = skcache_compaction(bk, bk->bid, 1, 0);
            Trace("compaction func. return %d <7>\n", chk);
 assert(bk->cup.num==bk->entry_use);
        }

    }
#endif

#if 0
    if(bk->frag_use >= bk->frag_hhwater || bk->frag_freelist==NULL) {
#if 1
        Trace("new_frag bk %p, used %4d/%d, p %p\n",
            bk, bk->frag_use, bk->frag_all, bk->frag_freelist);
#endif
        chk = skcache_compaction(bk, -7, 0, 1);
        Trace("compaction func. return %d <4a>\n", chk);
 assert(bk->cup.num==bk->entry_use);
		if(chk==0) {
        	chk = skcache_compaction(bk, -127, 1, 1);
        Trace("compaction func. return %d <4a>\n", chk);
 assert(bk->cup.num==bk->entry_use);
		}
    }
#endif





	if(bk->entry_freelist==NULL) {
		Error("no free element in entry free list of bk# %d <2v>",
			bk->bid, bk);

		return NULL;
	}

	if(bk->entry_use >= bk->entry_all) {
		Error("no free element in entry free list of bk# %d <2w>",
			bk->bid, bk);

		return NULL;
	}

    newone = bk->entry_freelist;
    bk->entry_freelist = newone->next;

#if 0
    newone->next = NULL;
	newone->ref = 0;
#endif
#if 1
	memset(newone, 0, sizeof(skcache_entry_t));
	newone->magic[0] = MAGIC_0;
	newone->magic[1] = MAGIC_1;
	newone->h1 = newone->h2 = newone->h3 = newone->h4 = -1;
	newone->issue_time = -1;
#endif

    bk->entry_use++;

#if 1
	if(bk->elem_freelist==NULL) {
		Trace("bk# %d, %p, elem last one\n", bk->bid, bk);
	}
	if(bk->entry_freelist==NULL) {
		Trace("bk# %d, %p, entry last one\n", bk->bid, bk);
	}
	
 assert(check_bkentry(bk, "nw4"));
#endif

	if(newone==NULL) {
		Error("skcache_pull_new_entry: return empty");
	}

    return newone;
}



int
skcache_push_idle_entry(bucket_t *bk, skcache_entry_t *xen)
{

#if 0
	Trace("idle_entry bk %p, xen %p, used %d/%d\n",
		bk, xen, bk->entry_use, bk->entry_all);
#endif

	if(bk==NULL) {
		Error("skcache_push_idle_entry: empty bucket");
		return -1;
	}
	if(xen==NULL) {
		Error("skcache_push_idle_entry: don't push NULL");
		return -1;
	}

#if 1
    if(bk->entry_use<=0) {
        Error("bk %p (%d), en %p, underflow entry", bk, bk->bid, xen);
		return -1;
    }
#endif

#if 0
	xen->ref = -1;
#endif
#if 1
	memset(xen, 0, sizeof(skcache_entry_t));
	xen->h1 = xen->h2 = xen->h3 = xen->h4 = -1;
	xen->ref = -1;
#endif

    xen->next = bk->entry_freelist;
    bk->entry_freelist = xen;

    bk->entry_use--;

#if 0
{
	void* pos;

	pos = sk_find(&bk->cup, (void*)xen, skcache_entry_cmp);
	Trace("verify. it is found in pos %p\n", pos);
}
#endif

    return 0;
}






long long
skcache_init(time_t xmin_date, int xbk_num, int xsk_level, int xsk_len,
        int xffl_num, int xffl_size)
{
    int i,j;
    int chk;
    long long sum;
    long long all;
    bucket_t *bk;
	time_t now;

	long long tmp;

#if 1
    Trace("skcache_init: min_date %d\n\
    bucket: number %d, level %d, length %d\n\
    frag: length %d, size %d\n",
        xmin_date, xbk_num, xsk_level, xsk_len, xffl_num, xffl_size);
#endif

    if(xmin_date>0) {
        min_date = xmin_date;
    }
    else {
#if 0
        min_date = time(NULL);
        Trace("set min_date as %d (current date/time)\n", min_date);
#endif
        Trace("set min_date as negative it is mean, use stored objects before boot\n");
    }

	now = min_date;

    sum = 0L;

    NIL = (skcache_entry_t*)malloc(sizeof(bucket_t));
    NIL->h1 = INT_MAX;
    NIL->h2 = INT_MAX;
    NIL->h3 = INT_MAX;

    sum += sizeof(bucket_t);

    skcache_bucket_base = (bucket_t*)malloc(sizeof(bucket_t)*xbk_num);
    if(skcache_bucket_base==NULL) {
        Error("not enough memory for skcache_bucket_base");
        skcache_bucket_num = 0;
        return -1;
    }
    sum += sizeof(bucket_t)*xbk_num;

    skcache_bucket_num = xbk_num;

	
    radix_h1 = skcache_bucket_num;
	radix_h2 = require_radix_h2;
	radix_h3 = require_radix_h3;

	swapout_token = swapout_token_init = require_swap_token_val;
	swapout_token_interval = require_swap_token_interval;
	disk_sweep_interval = require_disk_sweep_interval;
	dir_size_max = require_dir_size;


    if(radix_h2>RADIX_H2_MAX) {
        Error("radix_h2 (%d) is too large (max %d)\n",
            radix_h2, RADIX_H2_MAX);
    }
    dir_num = radix_h1*radix_h2*radix_h3;


#if 1
    Trace("radix of hashes = %d, %d, %d, %d\n",
        radix_h1, radix_h2, radix_h3, radix_h4);
    Trace("So, max. of directories is %d\n\n",
        dir_num);

	tmp = ((long long) dir_size_max * dir_num);
	Trace("Max size of directories is %lld\n", (long long) dir_size_max);
	Trace("So, actual capacity of cache on disk is %lld %.2fM %.2fG\n",
			tmp, fdM(tmp), fdG(tmp));

#endif

	if(dir_num<0) {
		Trace("dir_num %d is stange\n", dir_num);
		exit(4);
	}

    dir_sig = (int*)malloc(sizeof(int)*dir_num);
    sum += sizeof(int)*dir_num;

    dir_sz  = (int*)malloc(sizeof(int)*dir_num);
    sum += sizeof(int)*dir_num;

    for(i=0;i<dir_num;i++) {
        dir_sig[i] = -1;
        dir_sz[i]  = 0;
    }


    for(i=0;i<skcache_bucket_num;i++) {
#if 0
Trace("bk# %d %p\n", i, skcache_bucket_base[i]);
#endif

        chk = bk_new(&skcache_bucket_base[i], xsk_level, xsk_len, NIL);
        if(chk==0) {
            Error("initialize error in bucket# %d", i);
            return -2;
        }
        sum += chk;

        bk = &skcache_bucket_base[i];
		bk->bid = i;
		bk->own = '.';

        chk = alloc_cache_entry_freelist(&bk->entry_freelist, xsk_len+1);
        if(chk==0) {
            Error("initialize error in object entry elem_freelist# %d", i);
            return -2;
        }
        sum += chk;

        bk->entry_all = xsk_len+1;
        bk->entry_use = 0;
        bk->entry_hhwater   = 19*bk->entry_all/20;	/* 95% */
        bk->entry_hwater    = 6*bk->entry_all/10;	/* 60% */
        bk->entry_lwater    = 4*bk->entry_all/10;

        chk = alloc_object_frag_freelist(&bk->frag_freelist,
            xffl_num, xffl_size);
        if(chk==0) {
            Error("initialize error in object fragment elem_freelist# %d", i);
            return -2;
        }
        sum += chk;

        bk->frag_all = xffl_num;
        bk->frag_use = 0;
        /* 95% of all or under 16K of end */
        bk->frag_hhwater = MIN(19*bk->frag_all/20, xffl_num - K(16)/xffl_size);
        /* 50% of all or under 32K of end */
        bk->frag_hwater  = MIN(5*bk->frag_all/10, xffl_num - K(32)/xffl_size);
        bk->frag_lwater  = 4*bk->frag_all/10;

		bk->last_gc = min_date;
		bk->gc_push = 0;		/* never pushed */
        bk->dirty2 = -1;
        bk->dirty3 = -1;

    }

    Trace("skcache_init: allocate %lld byte(s) %dK %.2fM %.2fG\n",
        (long long) sum, (long)dK(sum), fdM(sum), fdG(sum));

    sim_stat_alloc_new(&swapload_time,  "swapload time [msec]",
        K(1),   swapload_time_count);
    sim_stat_alloc_new(&swapin_time,    "swapin time [msec]",
        K(1),   swapin_time_count);
    sim_stat_alloc_new(&swapout_time,   "swapout time [msec]",
        K(1),   swapout_time_count);
    sim_stat_alloc_new(&compaction_time,"compaction time [100msec]",
        K(2),   compaction_time_count);
    sim_stat_alloc_new(&checkdir_time,  "checkdir time [100msec]",
        K(2),   checkdir_time_count);
    sim_stat_alloc_new(&checkmem_time,  "checkmem time [100msec]",
        K(2),   checkmem_time_count);



    return sum;
}

